#
# License: See LICENSE.md file
# GitHub: https://github.com/Baekalfen/PyBoy
#

import io
import os.path
import platform
from pathlib import Path
from zipfile import ZipFile

import PIL
import pytest

from pyboy import PyBoy

if platform.python_implementation() == "PyPy":
    timeout = 15
else:
    timeout = 5

OVERWRITE_PNGS = False

saved_state = None


@pytest.mark.parametrize(
    "clean, rom",
    [
        ####################
        # These are meant to fail on DMG:
        # (False, "misc/boot_div-A.gb"),
        # (False, "misc/boot_div-cgb0.gb"),
        # (False, "misc/boot_div-cgbABCDE.gb"),
        # (False, "misc/boot_hwio-C.gb"),
        # (False, "misc/boot_regs-A.gb"),
        # (False, "misc/boot_regs-cgb.gb"),
        # (False, "misc/ppu/vblank_stat_intr-C.gb"),
        # (False, "misc/bits/unused_hwio-C.gb"),

        # (False, "utils/bootrom_dumper.gb"),
        # (False, "utils/dump_boot_hwio.gb"),
        (False, "manual-only/sprite_priority.gb"),
        (False, "acceptance/rapid_di_ei.gb"),
        (False, "acceptance/oam_dma_start.gb"),
        (False, "acceptance/boot_regs-dmgABC.gb"),
        (False, "acceptance/reti_timing.gb"),
        (False, "acceptance/call_timing.gb"),
        (False, "acceptance/reti_intr_timing.gb"),
        (False, "acceptance/boot_regs-mgb.gb"),
        (False, "acceptance/ei_sequence.gb"),
        (False, "acceptance/jp_timing.gb"),
        (False, "acceptance/ei_timing.gb"),
        (False, "acceptance/oam_dma_timing.gb"),
        (False, "acceptance/call_cc_timing2.gb"),
        (False, "acceptance/boot_div2-S.gb"),
        (False, "acceptance/halt_ime1_timing.gb"),
        (False, "acceptance/halt_ime1_timing2-GS.gb"),
        (False, "acceptance/timer/tima_reload.gb"),
        (False, "acceptance/timer/tma_write_reloading.gb"),
        (False, "acceptance/timer/tim10.gb"),
        (False, "acceptance/timer/tim00.gb"),
        (False, "acceptance/timer/tim11.gb"),
        (False, "acceptance/timer/tim01.gb"),
        (False, "acceptance/timer/tima_write_reloading.gb"),
        (False, "acceptance/timer/tim11_div_trigger.gb"),
        (False, "acceptance/timer/div_write.gb"),
        (False, "acceptance/timer/tim10_div_trigger.gb"),
        (False, "acceptance/timer/tim00_div_trigger.gb"),
        (False, "acceptance/timer/rapid_toggle.gb"),
        (False, "acceptance/timer/tim01_div_trigger.gb"),
        (False, "acceptance/boot_regs-sgb.gb"),
        (False, "acceptance/jp_cc_timing.gb"),
        (False, "acceptance/call_timing2.gb"),
        (False, "acceptance/ld_hl_sp_e_timing.gb"),
        (False, "acceptance/push_timing.gb"),
        (False, "acceptance/boot_hwio-dmg0.gb"),
        (False, "acceptance/rst_timing.gb"),
        (False, "acceptance/boot_hwio-S.gb"),
        (False, "acceptance/boot_div-dmgABCmgb.gb"),
        (False, "acceptance/bits/mem_oam.gb"),
        (False, "acceptance/bits/reg_f.gb"),
        (False, "acceptance/bits/unused_hwio-GS.gb"),
        (False, "acceptance/div_timing.gb"),
        (False, "acceptance/ret_cc_timing.gb"),
        (False, "acceptance/boot_regs-dmg0.gb"),
        (False, "acceptance/interrupts/ie_push.gb"),
        (False, "acceptance/boot_hwio-dmgABCmgb.gb"),
        (False, "acceptance/pop_timing.gb"),
        (False, "acceptance/ret_timing.gb"),
        (False, "acceptance/oam_dma_restart.gb"),
        (False, "acceptance/add_sp_e_timing.gb"),
        (False, "acceptance/oam_dma/sources-GS.gb"),
        (False, "acceptance/oam_dma/basic.gb"),
        (False, "acceptance/oam_dma/reg_read.gb"),
        (False, "acceptance/halt_ime0_nointr_timing.gb"),
        (False, "acceptance/ppu/vblank_stat_intr-GS.gb"),
        (False, "acceptance/ppu/intr_2_mode0_timing_sprites.gb"),
        (False, "acceptance/ppu/stat_irq_blocking.gb"),
        (False, "acceptance/ppu/intr_1_2_timing-GS.gb"),
        (False, "acceptance/ppu/intr_2_mode0_timing.gb"),
        (False, "acceptance/ppu/lcdon_write_timing-GS.gb"),
        (False, "acceptance/ppu/hblank_ly_scx_timing-GS.gb"),
        (False, "acceptance/ppu/intr_2_0_timing.gb"),
        (False, "acceptance/ppu/stat_lyc_onoff.gb"),
        (False, "acceptance/ppu/intr_2_mode3_timing.gb"),
        (False, "acceptance/ppu/lcdon_timing-GS.gb"),
        (False, "acceptance/ppu/intr_2_oam_ok_timing.gb"),
        (False, "acceptance/call_cc_timing.gb"),
        (False, "acceptance/halt_ime0_ei.gb"),
        (False, "acceptance/intr_timing.gb"),
        (False, "acceptance/instr/daa.gb"),
        (False, "acceptance/if_ie_registers.gb"),
        (False, "acceptance/di_timing-GS.gb"),
        (False, "acceptance/serial/boot_sclk_align-dmgABCmgb.gb"),
        (False, "acceptance/boot_regs-sgb2.gb"),
        (False, "acceptance/boot_div-S.gb"),
        (False, "acceptance/boot_div-dmg0.gb"),
        (True, "emulator-only/mbc5/rom_64Mb.gb"),
        (True, "emulator-only/mbc5/rom_1Mb.gb"),
        (True, "emulator-only/mbc5/rom_512kb.gb"),
        (True, "emulator-only/mbc5/rom_32Mb.gb"),
        (True, "emulator-only/mbc5/rom_2Mb.gb"),
        (True, "emulator-only/mbc5/rom_4Mb.gb"),
        (True, "emulator-only/mbc5/rom_8Mb.gb"),
        (True, "emulator-only/mbc5/rom_16Mb.gb"),
        # (True, "emulator-only/mbc2/rom_1Mb.gb"),
        # (True, "emulator-only/mbc2/ram.gb"),
        # (True, "emulator-only/mbc2/bits_unused.gb"),
        # (True, "emulator-only/mbc2/bits_ramg.gb"),
        # (True, "emulator-only/mbc2/rom_512kb.gb"),
        # (True, "emulator-only/mbc2/bits_romb.gb"),
        # (True, "emulator-only/mbc2/rom_2Mb.gb"),
        (True, "emulator-only/mbc1/rom_1Mb.gb"),
        (True, "emulator-only/mbc1/bits_bank2.gb"),
        (True, "emulator-only/mbc1/bits_ramg.gb"),
        (True, "emulator-only/mbc1/rom_512kb.gb"),
        (True, "emulator-only/mbc1/bits_mode.gb"),
        (True, "emulator-only/mbc1/ram_64kb.gb"),
        (True, "emulator-only/mbc1/bits_bank1.gb"),
        (True, "emulator-only/mbc1/rom_2Mb.gb"),
        (True, "emulator-only/mbc1/ram_256kb.gb"),
        (True, "emulator-only/mbc1/rom_4Mb.gb"),
        (True, "emulator-only/mbc1/multicart_rom_8Mb.gb"),
        (True, "emulator-only/mbc1/rom_8Mb.gb"),
        (True, "emulator-only/mbc1/rom_16Mb.gb"),
    ]
)
def test_mooneye(clean, rom, mooneye_dir, default_rom):
    global saved_state

    if saved_state is None:
        # HACK: We load any rom and load it until the last frame in the boot rom.
        # Then we save it, so we won't need to redo it.
        pyboy = PyBoy(default_rom, window_type="headless", cgb=False, sound_emulated=True)
        pyboy.set_emulation_speed(0)
        saved_state = io.BytesIO()
        for _ in range(59):
            pyboy.tick()
        pyboy.save_state(saved_state)
        pyboy.stop(save=False)

    pyboy = PyBoy(mooneye_dir + rom, window_type="headless", cgb=False)
    pyboy.set_emulation_speed(0)
    saved_state.seek(0)
    if clean:
        for _ in range(59):
            pyboy.tick()
    else:
        pyboy.load_state(saved_state)

    for _ in range(180 if "div_write" in rom else 40):
        pyboy.tick()

    png_path = Path(f"tests/test_results/mooneye/{rom}.png")
    image = pyboy.botsupport_manager().screen().screen_image()
    if OVERWRITE_PNGS:
        png_path.parents[0].mkdir(parents=True, exist_ok=True)
        image.save(png_path)
    else:
        old_image = PIL.Image.open(png_path)
        if "acceptance" in rom:
            # The registers are too volatile to depend on. We crop the top out, and only match the assertions.
            diff = PIL.ImageChops.difference(image.crop((0, 72, 160, 144)), old_image.crop((0, 72, 160, 144)))
        else:
            diff = PIL.ImageChops.difference(image, old_image)

        if diff.getbbox() and not os.environ.get("TEST_CI"):
            image.show()
            old_image.show()
            diff.show()
        assert not diff.getbbox(), f"Images are different! {rom}"

    pyboy.stop(save=False)



#
# License: See LICENSE.md file
# GitHub: https://github.com/Baekalfen/PyBoy
#

import os

import pytest
from tests.test_replay import replay

from .conftest import BOOTROM_FRAMES_UNTIL_LOGO

replay_file = "tests/replays/default_rom.replay"


def test_headless(default_rom):
    replay(default_rom, replay_file, "headless", bootrom_file=None, padding_frames=BOOTROM_FRAMES_UNTIL_LOGO)


def test_dummy(default_rom):
    replay(default_rom, replay_file, "dummy", bootrom_file=None, verify=False)


@pytest.mark.skipif(os.environ.get("TEST_NO_UI"), reason="Skipping test, as there is no UI")
def test_sdl2(default_rom):
    replay(default_rom, replay_file, "SDL2", bootrom_file=None, padding_frames=BOOTROM_FRAMES_UNTIL_LOGO)


@pytest.mark.skipif(os.environ.get("TEST_NO_UI"), reason="Skipping test, as there is no UI")
def test_opengl(default_rom):
    replay(default_rom, replay_file, "OpenGL", bootrom_file=None, padding_frames=BOOTROM_FRAMES_UNTIL_LOGO)



#
# License: See LICENSE.md file
# GitHub: https://github.com/Baekalfen/PyBoy
#

import cython

from cpython.array cimport array
from libc.stdint cimport uint8_t, uint16_t, uint32_t

from pyboy.logging.logging cimport Logger

from .base_mbc cimport BaseMBC


cdef Logger logger

@cython.locals(carttype=uint8_t, cart_name=basestring, cart_line=basestring)
cpdef BaseMBC load_cartridge(str) noexcept
cdef bint validate_checksum(uint8_t[:,:]) noexcept

@cython.locals(romdata=array, banksize=int)
cdef uint8_t[:, :] load_romfile(str) noexcept

cdef dict CARTRIDGE_TABLE
cdef dict EXTERNAL_RAM_TABLE



#
# License: See LICENSE.md file
# GitHub: https://github.com/Baekalfen/PyBoy
#
cimport cython
from libc.stdint cimport uint8_t

from pyboy.logging.logging cimport Logger
from pyboy.plugins.base_plugin cimport PyBoyGameWrapper


cdef Logger logger

cdef class GameWrapperPokemonGen1(PyBoyGameWrapper):
    pass



#
# License: See LICENSE.md file
# GitHub: https://github.com/Baekalfen/PyBoy

import os
import struct
import time

import pyboy
from pyboy import utils
from pyboy.utils import STATE_VERSION, IntIOWrapper

logger = pyboy.logging.get_logger(__name__)


class RTC:
    def __init__(self, filename):
        self.filename = filename + ".rtc"

        if not os.path.exists(self.filename):
            logger.info("No RTC file found. Skipping.")
        else:
            with open(self.filename, "rb") as f:
                self.load_state(IntIOWrapper(f), STATE_VERSION)

        self.latch_enabled = False

        self.timezero = time.time()

        self.sec_latch = 0
        self.min_latch = 0
        self.hour_latch = 0
        self.day_latch_low = 0
        self.day_latch_high = 0
        self.day_carry = 0
        self.halt = 0

    def stop(self):
        with open(self.filename, "wb") as f:
            self.save_state(IntIOWrapper(f))

    def save_state(self, f):
        for b in struct.pack("f", self.timezero):
            f.write(b)
        f.write(self.halt)
        f.write(self.day_carry)

    def load_state(self, f, state_version):
        self.timezero = int(struct.unpack("f", bytes([f.read() for _ in range(4)]))[0])
        self.halt = f.read()
        self.day_carry = f.read()

    def latch_rtc(self):
        t = time.time() - self.timezero
        self.sec_latch = int(t % 60)
        self.min_latch = int((t//60) % 60)
        self.hour_latch = int((t//3600) % 24)
        days = int(t // 3600 // 24)
        self.day_latch_low = days & 0xFF
        self.day_latch_high = days >> 8

        if self.day_latch_high > 1:
            self.day_carry = 1
            self.day_latch_high &= 0b1
            # Add 0x200 (512) days to "reset" the day counter to zero
            self.timezero += 0x200 * 3600 * 24

    def writecommand(self, value):
        if value == 0x00:
            self.latch_enabled = False
        elif value == 0x01:
            if not self.latch_enabled:
                self.latch_rtc()
            self.latch_enabled = True
        else:
            logger.warning("Invalid RTC command: %0.2x", value)

    def getregister(self, register):
        if not self.latch_enabled:
            logger.debug("RTC: Get register, but nothing is latched! 0x%0.2x", register)

        if register == 0x08:
            return self.sec_latch
        elif register == 0x09:
            return self.min_latch
        elif register == 0x0A:
            return self.hour_latch
        elif register == 0x0B:
            return self.day_latch_low
        elif register == 0x0C:
            day_high = self.day_latch_high & 0b1
            halt = self.halt << 6
            day_carry = self.day_carry << 7
            return day_high + halt + day_carry
        else:
            logger.warning("Invalid RTC register: %0.4x", register)

    def setregister(self, register, value):
        if not self.latch_enabled:
            logger.debug("RTC: Set register, but nothing is latched! 0x%0.4x, 0x%0.2x", register, value)

        t = time.time() - self.timezero
        if register == 0x08:
            # TODO: What happens, when these value are larger than allowed?
            self.timezero = self.timezero - (t%60) - value
        elif register == 0x09:
            self.timezero = self.timezero - (t//60%60) - value
        elif register == 0x0A:
            self.timezero = self.timezero - (t//3600%24) - value
        elif register == 0x0B:
            self.timezero = self.timezero - (t//3600//24) - value
        elif register == 0x0C:
            day_high = value & 0b1
            halt = (value & 0b1000000) >> 6
            day_carry = (value & 0b10000000) >> 7

            self.halt = halt
            if self.halt == 0:
                pass # TODO: Start the timer
            else:
                logger.warning("Stopping RTC is not implemented!")

            self.timezero = self.timezero - (t//3600//24) - (day_high << 8)
            self.day_carry = day_carry
        else:
            logger.warning("Invalid RTC register: %0.4x %0.2x", register, value)



#
# License: See LICENSE.md file
# GitHub: https://github.com/Baekalfen/PyBoy
#

import os.path
from pathlib import Path

import PIL

from pyboy import PyBoy

OVERWRITE_PNGS = False


# https://github.com/mattcurrie/cgb-acid2
def test_cgb_acid(cgb_acid_file):
    pyboy = PyBoy(cgb_acid_file, window_type="headless")
    pyboy.set_emulation_speed(0)
    for _ in range(59):
        pyboy.tick()

    for _ in range(25):
        pyboy.tick()

    png_path = Path(f"tests/test_results/{os.path.basename(cgb_acid_file)}.png")
    image = pyboy.botsupport_manager().screen().screen_image()
    if OVERWRITE_PNGS:
        png_path.parents[0].mkdir(parents=True, exist_ok=True)
        image.save(png_path)
    else:
        old_image = PIL.Image.open(png_path)
        diff = PIL.ImageChops.difference(image, old_image)
        if diff.getbbox() and not os.environ.get("TEST_CI"):
            image.show()
            old_image.show()
            diff.show()
        assert not diff.getbbox(), f"Images are different! {cgb_acid_file}"

    pyboy.stop(save=False)



#
# License: See LICENSE.md file
# GitHub: https://github.com/Baekalfen/PyBoy
#

import io
import json
import os.path
import platform
import sys
import time
from pathlib import Path

import pytest

from pyboy import PyBoy

OVERWRITE_JSON = False

blargg_json = "tests/test_results/blargg.json"


def run_rom(rom):
    pyboy = PyBoy(str(rom), window_type="dummy", cgb="cgb" in rom, sound_emulated=True)
    pyboy.set_emulation_speed(0)
    t = time.time()
    result = ""
    while not pyboy.tick():
        b = pyboy._serial()
        if b != "":
            result += b
            t = time.time()

        if pyboy._is_cpu_stuck():
            break

    for _ in range(10):
        pyboy.tick()
    pyboy.stop(save=False)
    result += pyboy._serial() # Getting the absolute last. Some times the tests says "Failed X tests".
    if result == "":
        n = 0
        while True:
            char = pyboy.get_memory_value(0xA004 + n)
            if char != 0:
                result += chr(char)
            n += 1

            if n > 250:
                break
    return result


@pytest.mark.parametrize(
    "test_rom", [
        "cgb_sound/cgb_sound.gb",
        "cgb_sound/rom_singles/01-registers.gb",
        "cgb_sound/rom_singles/02-len ctr.gb",
        "cgb_sound/rom_singles/03-trigger.gb",
        "cgb_sound/rom_singles/04-sweep.gb",
        "cgb_sound/rom_singles/05-sweep details.gb",
        "cgb_sound/rom_singles/06-overflow on trigger.gb",
        "cgb_sound/rom_singles/07-len sweep period sync.gb",
        "cgb_sound/rom_singles/08-len ctr during power.gb",
        "cgb_sound/rom_singles/09-wave read while on.gb",
        "cgb_sound/rom_singles/10-wave trigger while on.gb",
        "cgb_sound/rom_singles/11-regs after power.gb",
        "cgb_sound/rom_singles/12-wave.gb",
        "cpu_instrs/cpu_instrs.gb",
        "cpu_instrs/individual/01-special.gb",
        "cpu_instrs/individual/02-interrupts.gb",
        "cpu_instrs/individual/03-op sp,hl.gb",
        "cpu_instrs/individual/04-op r,imm.gb",
        "cpu_instrs/individual/05-op rp.gb",
        "cpu_instrs/individual/06-ld r,r.gb",
        "cpu_instrs/individual/07-jr,jp,call,ret,rst.gb",
        "cpu_instrs/individual/08-misc instrs.gb",
        "cpu_instrs/individual/09-op r,r.gb",
        "cpu_instrs/individual/10-bit ops.gb",
        "cpu_instrs/individual/11-op a,(hl).gb",
        "dmg_sound/dmg_sound.gb",
        "dmg_sound/rom_singles/01-registers.gb",
        "dmg_sound/rom_singles/02-len ctr.gb",
        "dmg_sound/rom_singles/03-trigger.gb",
        "dmg_sound/rom_singles/04-sweep.gb",
        "dmg_sound/rom_singles/05-sweep details.gb",
        "dmg_sound/rom_singles/06-overflow on trigger.gb",
        "dmg_sound/rom_singles/07-len sweep period sync.gb",
        "dmg_sound/rom_singles/08-len ctr during power.gb",
        "dmg_sound/rom_singles/09-wave read while on.gb",
        "dmg_sound/rom_singles/10-wave trigger while on.gb",
        "dmg_sound/rom_singles/11-regs after power.gb",
        "dmg_sound/rom_singles/12-wave write while on.gb",
        "instr_timing/instr_timing.gb",
        "interrupt_time/interrupt_time.gb",
        "mem_timing/individual/01-read_timing.gb",
        "mem_timing/individual/02-write_timing.gb",
        "mem_timing/individual/03-modify_timing.gb",
        "mem_timing/mem_timing.gb",
        "mem_timing-2/mem_timing.gb",
        "mem_timing-2/rom_singles/01-read_timing.gb",
        "mem_timing-2/rom_singles/02-write_timing.gb",
        "mem_timing-2/rom_singles/03-modify_timing.gb",
        "oam_bug/oam_bug.gb",
        "oam_bug/rom_singles/1-lcd_sync.gb",
        "oam_bug/rom_singles/2-causes.gb",
        "oam_bug/rom_singles/3-non_causes.gb",
        "oam_bug/rom_singles/4-scanline_timing.gb",
        "oam_bug/rom_singles/5-timing_bug.gb",
        "oam_bug/rom_singles/6-timing_no_bug.gb",
        "oam_bug/rom_singles/7-timing_effect.gb",
        "oam_bug/rom_singles/8-instr_effect.gb",
    ]
)
def test_blarggs(test_rom, blargg_dir):
    rom = str(blargg_dir / Path(test_rom))
    result = run_rom(rom)

    if os.path.isfile(blargg_json):
        with open(blargg_json, "r") as f:
            old_blargg = json.load(f)
    else:
        old_blargg = None

    rom = rom.replace("\\", "/") # Fix Windows lookup
    rom = rom.replace("test_roms/", "")
    if OVERWRITE_JSON:
        with open(blargg_json, "w") as f:
            old_blargg[rom] = result
            json.dump(old_blargg, f, indent=4)

        blargg_md = "../PyBoy.wiki/blargg.md"
        if os.path.isdir(blargg_md):
            with open(blargg_md, "w") as f:
                f.write("# Test results for Blargg's test ROMs\n")
                f.write("|ROM|Result|\n")
                f.write("|---|---|\n")
                for (rom, _), res in zip(test_roms, old_blargg):
                    f.write("|%s|%s|\n" % (rom, res.replace("\n", " ").rstrip(":")))
    else:
        assert old_blargg[rom] == result, f"Outputs don't match for {rom}"



#
# License: See LICENSE file
# GitHub: https://github.com/Baekalfen/PyBoy
#
# Based on specs from:
# http://gbdev.gg8.se/wiki/articles/Sound_Controller
# http://gbdev.gg8.se/wiki/articles/Gameboy_sound_hardware
# http://www.devrs.com/gb/files/hosted/GBSOUND.txt

from array import array
from ctypes import c_void_p

from pyboy import utils

try:
    import sdl2
except ImportError:
    sdl2 = None

import pyboy

logger = pyboy.logging.get_logger(__name__)

SOUND_DESYNC_THRESHOLD = 5
CPU_FREQ = 4213440 # hz


class Sound:
    def __init__(self, enabled, emulate):
        self.enabled = enabled and (sdl2 is not None)
        self.emulate = emulate or enabled # Just emulate registers etc.
        if self.enabled:
            # Initialization is handled in the windows, otherwise we'd need this
            if sdl2.SDL_Init(sdl2.SDL_INIT_AUDIO) >= 0:
                # Open audio device
                spec_want = sdl2.SDL_AudioSpec(32768, sdl2.AUDIO_S8, 2, 64)
                spec_have = sdl2.SDL_AudioSpec(0, 0, 0, 0)
                self.device = sdl2.SDL_OpenAudioDevice(None, 0, spec_want, spec_have, 0)

                if self.device > 1:
                    # Start playback (move out of __init__ if needed, maybe for headless)
                    sdl2.SDL_PauseAudioDevice(self.device, 0)

                    self.sample_rate = spec_have.freq
                    self.sampleclocks = CPU_FREQ // self.sample_rate
                else:
                    # TODO: Refactor SDL2 out of sound.py
                    logger.error("SDL_OpenAudioDevice failed: %s", sdl2.SDL_GetError().decode())
                    self.enabled = False # We will continue with emulation
            else:
                # TODO: Refactor SDL2 out of sound.py
                logger.error("SDL_Init audio failed: %s", sdl2.SDL_GetError().decode())
                self.enabled = False # We will continue with emulation

        if not self.enabled:
            self.sample_rate = 32768
            self.sampleclocks = CPU_FREQ // self.sample_rate

        self.audiobuffer = array("b", [0] * 4096) # Over 2 frames
        self.audiobuffer_p = c_void_p(self.audiobuffer.buffer_info()[0])

        self.clock = 0

        self.poweron = True

        self.sweepchannel = SweepChannel()
        self.tonechannel = ToneChannel()
        self.wavechannel = WaveChannel()
        self.noisechannel = NoiseChannel()

        self.leftnoise = False
        self.leftwave = False
        self.lefttone = False
        self.leftsweep = False
        self.rightnoise = False
        self.rightwave = False
        self.righttone = False
        self.rightsweep = False

    def get(self, offset):
        if not self.emulate:
            return 0
        self.sync()
        if offset < 20:
            i = offset // 5
            if i == 0:
                return self.sweepchannel.getreg(offset % 5)
            elif i == 1:
                return self.tonechannel.getreg(offset % 5)
            elif i == 2:
                return self.wavechannel.getreg(offset % 5)
            elif i == 3:
                return self.noisechannel.getreg(offset % 5)
        elif offset == 20: # Control register NR50: Vin enable and volume -- not implemented
            return 0
        elif offset == 21: # Control register NR51: Channel stereo enable/panning
            return ((0x80 if self.leftnoise else 0) | (0x40 if self.leftwave else 0) | (0x20 if self.lefttone else 0) |
                    (0x10 if self.leftsweep else 0) | (0x08 if self.rightnoise else 0) |
                    (0x04 if self.rightwave else 0) | (0x02 if self.righttone else 0) |
                    (0x01 if self.rightsweep else 0))
        elif offset == 22: # Control register NR52: Sound/channel enable
            return 0x70 | ((0x80 if self.poweron else 0) | (0x08 if self.noisechannel.enable else 0) |
                           (0x04 if self.wavechannel.enable else 0) | (0x02 if self.tonechannel.enable else 0) |
                           (0x01 if self.sweepchannel.enable else 0))
        elif offset < 32: # Unused registers, read as 0xFF
            return 0xFF
        elif offset < 48: # Wave Table
            return self.wavechannel.getwavebyte(offset - 32)
        else:
            raise IndexError(f"Attempted to read register {offset} in sound memory")

    def set(self, offset, value):
        if not self.emulate:
            return
        self.sync()
        if offset < 20 and self.poweron:
            i = offset // 5
            if i == 0:
                self.sweepchannel.setreg(offset % 5, value)
            elif i == 1:
                self.tonechannel.setreg(offset % 5, value)
            elif i == 2:
                self.wavechannel.setreg(offset % 5, value)
            elif i == 3:
                self.noisechannel.setreg(offset % 5, value)
        elif offset == 20 and self.poweron: # Control register NR50: Vin enable and volume -- not implemented
            return
        elif offset == 21 and self.poweron: # Control register NR51: Channel stereo enable/panning
            self.leftnoise = value & 0x80
            self.leftwave = value & 0x40
            self.lefttone = value & 0x20
            self.leftsweep = value & 0x10
            self.rightnoise = value & 0x08
            self.rightwave = value & 0x04
            self.righttone = value & 0x02
            self.rightsweep = value & 0x01
            return
        elif offset == 22: # Control register NR52: Sound on/off
            if value & 0x80 == 0: # Sound power off
                for n in range(22):
                    self.set(n, 0)
                self.poweron = False
            else:
                self.poweron = True
            return
        elif offset < 32: # Unused registers, unwritable?
            return
        elif offset < 48: # Wave Table
            self.wavechannel.setwavebyte(offset - 32, value)
        else:
            raise IndexError(f"Attempted to write register {offset} in sound memory")

    def sync(self):
        """Run the audio for the number of clock cycles stored in self.clock"""
        if not self.emulate:
            return

        nsamples = self.clock // self.sampleclocks

        for i in range(min(2048, nsamples)):
            if self.poweron:
                self.sweepchannel.run(self.sampleclocks)
                self.tonechannel.run(self.sampleclocks)
                self.wavechannel.run(self.sampleclocks)
                self.noisechannel.run(self.sampleclocks)
                sample = ((self.sweepchannel.sample() if self.leftsweep else 0) +
                          (self.tonechannel.sample() if self.lefttone else 0) +
                          (self.wavechannel.sample() if self.leftwave else 0) +
                          (self.noisechannel.sample() if self.leftnoise else 0))
                self.audiobuffer[2 * i] = min(max(sample, 0), 127)
                sample = ((self.sweepchannel.sample() if self.rightsweep else 0) +
                          (self.tonechannel.sample() if self.righttone else 0) +
                          (self.wavechannel.sample() if self.rightwave else 0) +
                          (self.noisechannel.sample() if self.rightnoise else 0))
                self.audiobuffer[2*i + 1] = min(max(sample, 0), 127)
                self.clock -= self.sampleclocks
            else:
                self.audiobuffer[2 * i] = 0
                self.audiobuffer[2*i + 1] = 0

        if self.enabled:
            self.enqueue_sound(nsamples)
        self.clock %= self.sampleclocks

    def enqueue_sound(self, nsamples):
        # Clear queue, if we are behind
        queued_time = sdl2.SDL_GetQueuedAudioSize(self.device)
        samples_per_frame = (self.sample_rate / 60) * 2 # Data of 1 frame's worth (60) in stereo (2)
        if queued_time > samples_per_frame * SOUND_DESYNC_THRESHOLD:
            sdl2.SDL_ClearQueuedAudio(self.device)

        sdl2.SDL_QueueAudio(self.device, self.audiobuffer_p, 2 * nsamples)

    def stop(self):
        if self.enabled:
            sdl2.SDL_CloseAudioDevice(self.device)

    def save_state(self, file):
        pass

    def load_state(self, file, state_version):
        pass


class ToneChannel:
    """Second sound channel--simple square wave, no sweep"""
    def __init__(self):
        # Shape of square waves at different duty cycles
        # These could ostensibly be replaced with other waves for fun experiments
        self.wavetables = [
            [0, 0, 0, 0, 0, 0, 0, 1], # 12.5% Duty cycle square
            [1, 0, 0, 0, 0, 0, 0, 1], # 25%
            [1, 0, 0, 0, 0, 1, 1, 1], # 50%
            [0, 1, 1, 1, 1, 1, 1, 0]
        ] # 75% (25% inverted)

        # Register values (abbreviated to keep track of what's external)
        # Register 0 is unused in the non-sweep tone channel
        self.wavsel = 0 # Register 1 bits 7-6: wave table selection (duty cycle)
        self.sndlen = 0 # Register 1 bits 5-0: time to play sound before stop (64-x)
        self.envini = 0 # Register 2 bits 7-4: volume envelope initial volume
        self.envdir = 0 # Register 2 bit 3: volume envelope change direction (0: decrease)
        self.envper = 0 # Register 2 bits 2-0: volume envelope period (0: disabled)
        self.sndper = 0 # Register 4 bits 2-0 MSB + register 3 all: period of tone ("frequency" on gg8 wiki)
        # Register 4 bit 7: Write-only trigger bit. Process immediately.
        self.uselen = 0 # Register 4 bit 6: enable/disable sound length timer in reg 1 (0: continuous)

        # Internal values
        self.enable = False # Enable flag, turned on by trigger bit and off by length timer
        self.lengthtimer = 64 # Length timer, counts down to disable channel automatically
        self.periodtimer = 0 # Period timer, counts down to signal change in wave frame
        self.envelopetimer = 0 # Volume envelope timer, counts down to signal change in volume
        self.period = 4 # Calculated copy of period, 4 * (2048 - sndper)
        self.waveframe = 0 # Wave frame index into wave table entries
        self.frametimer = 0x2000 # Frame sequencer timer, underflows to signal change in frame sequences
        self.frame = 0 # Frame sequencer value, generates clocks for length/envelope/(sweep)
        self.volume = 0 # Current volume level, modulated by envelope

    def getreg(self, reg):
        if reg == 0:
            return 0
        elif reg == 1:
            return self.wavsel << 6 # Other bits are write-only
        elif reg == 2:
            return self.envini << 4 | self.envdir << 3 | self.envper
        elif reg == 3:
            return 0 # Write-only register?
        elif reg == 4:
            return self.uselen << 6 # Other bits are write-only
        else:
            raise IndexError("Attempt to read register {} in ToneChannel".format(reg))

    def setreg(self, reg, val):
        if reg == 0:
            return
        elif reg == 1:
            self.wavsel = (val >> 6) & 0x03
            self.sndlen = val & 0x1F
            self.lengthtimer = 64 - self.sndlen
        elif reg == 2:
            self.envini = (val >> 4) & 0x0F
            self.envdir = (val >> 3) & 0x01
            self.envper = val & 0x07
            if self.envini == 0 and self.envdir == 0:
                self.enable = False
        elif reg == 3:
            self.sndper = (self.sndper & 0x700) + val # Is this ever written solo?
            self.period = 4 * (0x800 - self.sndper)
        elif reg == 4:
            self.uselen = (val >> 6) & 0x01
            self.sndper = ((val << 8) & 0x0700) + (self.sndper & 0xFF)
            self.period = 4 * (0x800 - self.sndper)
            if val & 0x80:
                self.trigger() # Sync is called first in Sound.set so it's okay to trigger immediately
        else:
            raise IndexError("Attempt to write register {} in ToneChannel".format(reg))

    def run(self, clocks):
        """Advances time to sync with system state.

        Doesn't generate samples, but we assume that it is called at
        the sample rate or faster, so this might not be not prepared
        to handle high values for 'clocks'.

        """
        self.periodtimer -= clocks
        while self.periodtimer <= 0:
            self.periodtimer += self.period
            self.waveframe = (self.waveframe + 1) % 8

        self.frametimer -= clocks
        while self.frametimer <= 0:
            self.frametimer += 0x2000
            self.tickframe()

    def tickframe(self):
        self.frame = (self.frame + 1) % 8
        # Clock length timer on 0, 2, 4, 6
        if self.uselen and self.frame & 1 == 0 and self.lengthtimer > 0:
            self.lengthtimer -= 1
            if self.lengthtimer == 0:
                self.enable = False
        # Clock envelope timer on 7
        if self.frame == 7 and self.envelopetimer != 0:
            self.envelopetimer -= 1
            if self.envelopetimer == 0:
                newvolume = self.volume + (self.envdir or -1)
                if newvolume < 0 or newvolume > 15:
                    self.envelopetimer = 0
                else:
                    self.envelopetimer = self.envper
                    self.volume = newvolume
                # Note that setting envelopetimer to 0 disables it

    def sample(self):
        return self.volume * self.wavetables[self.wavsel][self.waveframe] if self.enable else 0

    def trigger(self):
        self.enable = True
        self.lengthtimer = self.lengthtimer or 64
        self.periodtimer = self.period
        self.envelopetimer = self.envper
        self.volume = self.envini
        # TODO: If channel DAC is off (NRx2 & 0xF8 == 0) then this
        #   will be undone and the channel immediately disabled.
        #   Probably need a new DAC power state/variable.
        # For now:
        if self.envper == 0 and self.envini == 0:
            self.enable = False


class SweepChannel(ToneChannel):
    def __init__(self):
        ToneChannel.__init__(self)

        # Register Values
        self.swpper = 0 # Register 0 bits 6-4: Sweep period
        self.swpdir = 0 # Register 0 bit 3: Sweep direction (0: increase)
        self.swpmag = 0 # Register 0 bits 2-0: Sweep size as a bit shift

        # Internal Values
        self.sweeptimer = 0 # Sweep timer, counts down to shift pitch
        self.sweepenable = False # Internal sweep enable flag
        self.shadow = 0 # Shadow copy of period register for ignoring writes to sndper

    def getreg(self, reg):
        if reg == 0:
            return self.swpper << 4 | self.swpdir << 3 | self.swpmag
        else:
            return ToneChannel.getreg(self, reg)

    def setreg(self, reg, val):
        if reg == 0:
            self.swpper = val >> 4 & 0x07
            self.swpdir = val >> 3 & 0x01
            self.swpmag = val & 0x07
        else:
            ToneChannel.setreg(self, reg, val)

    # run() is the same as in ToneChannel, so we only override tickframe
    def tickframe(self):
        ToneChannel.tickframe(self)
        # Clock sweep timer on 2 and 6
        if self.sweepenable and self.swpper and self.frame & 3 == 2:
            self.sweeptimer -= 1
            if self.sweeptimer == 0:
                if self.sweep(True):
                    self.sweeptimer = self.swpper
                    self.sweep(False)

    def trigger(self):
        ToneChannel.trigger(self)
        self.shadow = self.sndper
        self.sweeptimer = self.swpper
        self.sweepenable = True if (self.swpper or self.swpmag) else False
        if self.swpmag:
            self.sweep(False)

    def sweep(self, save):
        if self.swpdir == 0:
            newper = self.shadow + (self.shadow >> self.swpmag)
        else:
            newper = self.shadow - (self.shadow >> self.swpmag)
        if newper >= 0x800:
            self.enable = False
            return False
        elif save and self.swpmag:
            self.sndper = self.shadow = newper
            self.period = 4 * (0x800 - self.sndper)
            return True


class WaveChannel:
    """Third sound channel--sample-based playback"""
    def __init__(self):
        # Memory for wave sample
        self.wavetable = array("B", [0xFF] * 16)

        # Register values (abbreviated to keep track of what's external)
        # Register 0 is unused in the wave channel
        self.dacpow = 0 # Register 0 bit 7: DAC Power, enable playback
        self.sndlen = 0 # Register 1 bits 7-0: time to play sound before stop (256-x)
        self.volreg = 0 # Register 2 bits 6-5: volume code
        self.sndper = 0 # Register 4 bits 2-0 MSB + register 3 all: period of tone ("frequency" on gg8 wiki)
        # Register 4 bit 7: Write-only trigger bit. Process immediately.
        self.uselen = 0 # Register 4 bit 6: enable/disable sound length timer in reg 1 (0: continuous)

        # Internal values
        self.enable = False # Enable flag, turned on by trigger bit and off by length timer
        self.lengthtimer = 256 # Length timer, counts down to disable channel automatically
        self.periodtimer = 0 # Period timer, counts down to signal change in wave frame
        self.period = 4 # Calculated copy of period, 4 * (2048 - sndper)
        self.waveframe = 0 # Wave frame index into wave table entries
        self.frametimer = 0x2000 # Frame sequencer timer, underflows to signal change in frame sequences
        self.frame = 0 # Frame sequencer value, generates clocks for length/envelope/(sweep)
        self.volumeshift = 0 # Bitshift for volume, set by volreg

    def getreg(self, reg):
        # https://gbdev.gg8.se/wiki/articles/Gameboy_sound_hardware#Register_Reading
        if reg == 0:
            return self.dacpow << 7 | 0x7F
        elif reg == 1:
            return 0xFF
        elif reg == 2:
            return self.volreg << 5 | 0x9F
        elif reg == 3:
            return 0xFF
        elif reg == 4:
            return self.uselen << 6 | 0xBF
        else:
            raise IndexError("Attempt to read register {} in ToneChannel".format(reg))

    def setreg(self, reg, val):
        if reg == 0:
            self.dacpow = val >> 7 & 0x01
            if self.dacpow == 0:
                self.enable = False
        elif reg == 1:
            self.sndlen = val
            self.lengthtimer = 256 - self.sndlen
        elif reg == 2:
            self.volreg = val >> 5 & 0x03
            self.volumeshift = self.volreg - 1 if self.volreg > 0 else 4
        elif reg == 3:
            self.sndper = (self.sndper & 0x700) + val # Is this ever written solo?
            self.period = 2 * (0x800 - self.sndper)
        elif reg == 4:
            self.uselen = val >> 6 & 0x01
            self.sndper = (val << 8 & 0x0700) + (self.sndper & 0xFF)
            self.period = 2 * (0x800 - self.sndper)
            if val & 0x80:
                self.trigger() # Sync is called first in Sound.set so it's okay to trigger immediately
        else:
            raise IndexError("Attempt to write register {} in WaveChannel".format(reg))

    def getwavebyte(self, offset):
        if self.dacpow:
            return self.wavetable[self.waveframe % 16]
        else:
            return self.wavetable[offset]

    def setwavebyte(self, offset, value):
        # In GBA, a write is ignored while the channel is running.
        # Otherwise, it usually goes at the current frame byte.
        if self.dacpow:
            self.wavetable[self.waveframe % 16] = value
        else:
            self.wavetable[offset] = value

    def run(self, clocks):
        """Advances time to sync with system state."""
        self.periodtimer -= clocks
        while self.periodtimer <= 0:
            self.periodtimer += self.period
            self.waveframe += 1
            self.waveframe %= 32

        self.frametimer -= clocks
        while self.frametimer <= 0:
            self.frametimer += 0x2000
            self.tickframe()

    def tickframe(self):
        self.frame = (self.frame + 1) % 8
        # Clock length timer on 0, 2, 4, 6
        if self.uselen and self.frame & 1 == 0 and self.lengthtimer > 0:
            self.lengthtimer -= 1
            if self.lengthtimer == 0:
                self.enable = False

    def sample(self):
        if self.enable and self.dacpow:
            sample = self.wavetable[self.waveframe // 2] >> (0 if self.waveframe % 2 else 4) & 0x0F
            return sample >> self.volumeshift
        else:
            return 0

    def trigger(self):
        self.enable = True if self.dacpow else False
        self.lengthtimer = self.lengthtimer or 256
        self.periodtimer = self.period


class NoiseChannel:
    """Fourth sound channel--white noise generator"""
    def __init__(self):
        self.DIVTABLE = (8, 16, 32, 48, 64, 80, 96, 112)

        # Register values (abbreviated to keep track of what's external)
        # Register 0 is unused in the noise channel
        self.sndlen = 0 # Register 1 bits 5-0: time to play sound before stop (64-x)
        self.envini = 0 # Register 2 bits 7-4: volume envelope initial volume
        self.envdir = 0 # Register 2 bit 3: volume envelope change direction (0: decrease)
        self.envper = 0 # Register 2 bits 2-0: volume envelope period (0: disabled)
        self.clkpow = 0 # Register 3 bits 7-4: lfsr clock shift
        self.regwid = 0 # Register 3 bit 3: lfsr bit width (0: 15, 1: 7)
        self.clkdiv = 0 # Register 3 bits 2-0: base divider for lfsr clock
        # Register 4 bit 7: Write-only trigger bit. Process immediately.
        self.uselen = 0 # Register 4 bit 6: enable/disable sound length timer in reg 1 (0: continuous)

        # Internal values
        self.enable = False # Enable flag, turned on by trigger bit and off by length timer
        self.lengthtimer = 64 # Length timer, counts down to disable channel automatically
        self.periodtimer = 0 # Period timer, counts down to signal change in wave frame
        self.envelopetimer = 0 # Volume envelope timer, counts down to signal change in volume
        self.period = 8 # Calculated copy of period, 8 << 0
        self.shiftregister = 1 # Internal shift register value
        self.lfsrfeed = 0x4000 # Bit mask for inserting feedback in shift register
        self.frametimer = 0x2000 # Frame sequencer timer, underflows to signal change in frame sequences
        self.frame = 0 # Frame sequencer value, generates clocks for length/envelope/(sweep)
        self.volume = 0 # Current volume level, modulated by envelope

    def getreg(self, reg):
        if reg == 0:
            return 0xFF
        elif reg == 1:
            return 0xFF
        elif reg == 2:
            return self.envini << 4 | self.envdir << 3 | self.envper
        elif reg == 3:
            return self.clkpow << 4 | self.regwid << 3 | self.clkdiv
        elif reg == 4:
            return self.uselen << 6 | 0xBF
        else:
            raise IndexError("Attempt to read register {} in NoiseChannel".format(reg))

    def setreg(self, reg, val):
        if reg == 0:
            return
        elif reg == 1:
            self.sndlen = val & 0x1F
            self.lengthtimer = 64 - self.sndlen
        elif reg == 2:
            self.envini = val >> 4 & 0x0F
            self.envdir = val >> 3 & 0x01
            self.envper = val & 0x07
            if self.envini == 0 and self.envdir == 0:
                self.enable = False
        elif reg == 3:
            self.clkpow = val >> 4 & 0x0F
            self.regwid = val >> 3 & 0x01
            self.clkdiv = val & 0x07
            self.period = self.DIVTABLE[self.clkdiv] << self.clkpow
            self.lfsrfeed = 0x4040 if self.regwid else 0x4000
        elif reg == 4:
            self.uselen = val >> 6 & 0x01
            if val & 0x80:
                self.trigger() # Sync is called first in Sound.set so it's okay to trigger immediately
        else:
            raise IndexError("Attempt to write register {} in ToneChannel".format(reg))

    def run(self, clocks):
        """Advances time to sync with system state."""
        self.periodtimer -= clocks
        while self.periodtimer <= 0:
            self.periodtimer += self.period
            # Advance shift register
            # This is good C, but terrible Python
            tap = self.shiftregister
            self.shiftregister >>= 1
            tap ^= self.shiftregister
            if tap & 0x01:
                self.shiftregister |= self.lfsrfeed
            else:
                self.shiftregister &= ~self.lfsrfeed

        self.frametimer -= clocks
        while self.frametimer <= 0:
            self.frametimer += 0x2000
            self.tickframe()

    def tickframe(self):
        self.frame = (self.frame + 1) % 8
        # Clock length timer on 0, 2, 4, 6
        if self.uselen and self.frame & 1 == 0 and self.lengthtimer > 0:
            self.lengthtimer -= 1
            if self.lengthtimer == 0:
                self.enable = False
        # Clock envelope timer on 7
        if self.frame == 7 and self.envelopetimer != 0:
            self.envelopetimer -= 1
            if self.envelopetimer == 0:
                newvolume = self.volume + (self.envdir or -1)
                if newvolume < 0 or newvolume > 15:
                    self.envelopetimer = 0
                else:
                    self.envelopetimer = self.envper
                    self.volume = newvolume
                # Note that setting envelopetimer to 0 disables it

    def sample(self):
        if self.enable:
            return self.volume if self.shiftregister & 0x01 == 0 else 0
        else:
            return 0

    def trigger(self):
        self.enable = True
        self.lengthtimer = self.lengthtimer or 64
        self.periodtimer = self.period
        self.envelopetimer = self.envper
        self.volume = self.envini
        self.shiftregister = 0x7FFF
        # TODO: tidy instead of double change variable
        if self.envper == 0 and self.envini == 0:
            self.enable = False



#
# License: See LICENSE.md file
# GitHub: https://github.com/Baekalfen/PyBoy
#
"""
The Game Boy uses tiles as the building block for all graphics on the screen. This base-class is used both for
`pyboy.botsupport.sprite.Sprite` and `pyboy.botsupport.tilemap.TileMap`, when refering to graphics.
"""

import numpy as np

import pyboy
from pyboy import utils

from .constants import LOW_TILEDATA, VRAM_OFFSET

logger = pyboy.logging.get_logger(__name__)

try:
    from PIL import Image
except ImportError:
    Image = None


class Tile:
    def __init__(self, mb, identifier):
        """
        The Game Boy uses tiles as the building block for all graphics on the screen. This base-class is used for
        `pyboy.botsupport.BotSupportManager.tile`, `pyboy.botsupport.sprite.Sprite` and `pyboy.botsupport.tilemap.TileMap`, when
        refering to graphics.

        This class is not meant to be instantiated by developers reading this documentation, but it will be created
        internally and returned by `pyboy.botsupport.sprite.Sprite.tiles` and
        `pyboy.botsupport.tilemap.TileMap.tile`.

        The data of this class is static, apart from the image data, which is loaded from the Game Boy's memory when
        needed. Beware that the graphics for the tile can change between each call to `pyboy.PyBoy.tick`.
        """
        self.mb = mb

        assert 0 <= identifier < 384, "Identifier out of range"

        self.data_address = LOW_TILEDATA + (16*identifier)
        """
        The tile data is defined in a specific area of the Game Boy. This function returns the address of the tile data
        corresponding to the tile identifier. It is advised to use `pyboy.botsupport.tile.Tile.image` or one of the
        other `image`-functions if you want to view the tile.

        You can read how the data is read in the
        [Pan Docs: VRAM Tile Data](http://bgb.bircd.org/pandocs.htm#vramtiledata).

        Returns
        -------
        int:
            address in VRAM where tile data starts
        """

        self.tile_identifier = (self.data_address - LOW_TILEDATA) // 16
        """
        The Game Boy has a slightly complicated indexing system for tiles. This identifier unifies the otherwise
        complicated indexing system on the Game Boy into a single range of 0-383 (both included).

        Returns
        -------
        int:
            Unique identifier for the tile
        """

        self.shape = (8, 8)
        """
        Tiles are always 8x8 pixels.

        Returns
        -------
        (int, int):
            The width and height of the tile.
        """

    def image(self):
        """
        Use this function to get an easy-to-use `PIL.Image` object of the tile. The image is 8x8 pixels in RGBA colors.

        Be aware, that the graphics for this tile can change between each call to `pyboy.PyBoy.tick`.

        Returns
        -------
        PIL.Image :
            Image of tile in 8x8 pixels and RGBA colors.
        """
        if Image is None:
            logger.error(f"{__name__}: Missing dependency \"Pillow\".")
            return None
        return Image.frombytes("RGBA", (8, 8), bytes(self.image_data()))

    def image_ndarray(self):
        """
        Use this function to get an easy-to-use `numpy.ndarray` object of the tile. The array has a shape of (8, 8, 4)
        and each value is of `numpy.uint8`. The values corresponds to and RGBA image of 8x8 pixels with each sub-color
        in a separate cell.

        Be aware, that the graphics for this tile can change between each call to `pyboy.PyBoy.tick`.

        Returns
        -------
        numpy.ndarray :
            Array of shape (8, 8, 4) with data type of `numpy.uint8`.
        """
        return np.asarray(self.image_data()).view(dtype=np.uint8).reshape(8, 8, 4)

    def image_data(self):
        """
        Use this function to get the raw tile data. The data is a `memoryview` corresponding to 8x8 pixels in RGBA
        colors.

        Be aware, that the graphics for this tile can change between each call to `pyboy.PyBoy.tick`.

        Returns
        -------
        memoryview :
            Image data of tile in 8x8 pixels and RGBA colors.
        """
        self.data = np.zeros((8, 8), dtype=np.uint32)
        for k in range(0, 16, 2): # 2 bytes for each line
            byte1 = self.mb.lcd.VRAM0[self.data_address + k - VRAM_OFFSET]
            byte2 = self.mb.lcd.VRAM0[self.data_address + k + 1 - VRAM_OFFSET]

            for x in range(8):
                colorcode = utils.color_code(byte1, byte2, 7 - x)
                # NOTE: ">> 8 | 0xFF000000" to keep compatibility with earlier code
                old_A_format = 0xFF000000
                self.data[k // 2][x] = self.mb.lcd.BGP.getcolor(colorcode) >> 8 | old_A_format

        return self.data

    def __eq__(self, other):
        return self.data_address == other.data_address

    def __repr__(self):
        return f"Tile: {self.tile_identifier}"



#
# License: See LICENSE.md file
# GitHub: https://github.com/Baekalfen/PyBoy
#
cimport cython
from libc.stdint cimport uint8_t

from pyboy.logging.logging cimport Logger
from pyboy.plugins.base_plugin cimport PyBoyGameWrapper


cdef Logger logger

cdef int ROWS, COLS


cdef class GameWrapperSuperMarioLand(PyBoyGameWrapper):
    cdef public tuple world
    cdef public int coins
    cdef public int lives_left
    cdef public int score
    cdef public int time_left
    cdef public int level_progress
    cdef public int _level_progress_max
    cdef public int fitness

    cpdef void start_game(self, timer_div=*, world_level=*, unlock_level_select=*) noexcept
    cpdef void reset_game(self, timer_div=*) noexcept
    cpdef set_lives_left(self, int) noexcept


#
# License: See LICENSE.md file
# GitHub: https://github.com/Baekalfen/PyBoy
#
cimport cython
from libc.stdint cimport uint8_t

from pyboy.logging.logging cimport Logger
from pyboy.plugins.base_plugin cimport PyBoyGameWrapper


cdef Logger logger

cdef int ROWS, COLS


cdef class GameWrapperKirbyDreamLand(PyBoyGameWrapper):
    cdef public int score
    cdef public int health
    cdef public int lives_left
    cdef public int fitness
    cdef public int _game_over


import platform

import pytest

is_pypy = platform.python_implementation() == "PyPy"

if is_pypy:
    from pyboy.plugins.rewind import DeltaFixedAllocBuffers, CompressedFixedAllocBuffers, FixedAllocBuffers, \
                                    FILL_VALUE, FIXED_BUFFER_SIZE


def write_bytes(buf, values):
    for v in values:
        buf.write(v % 0x100)


@pytest.mark.skipif(not is_pypy, reason="This test doesn't work in Cython")
class TestRewind:
    def test_all(self):
        for buf in [FixedAllocBuffers(), CompressedFixedAllocBuffers(), DeltaFixedAllocBuffers()]:
            A = [1] * 16
            B = [2] * 16
            C = [4] * 16
            # A = [x % 0x100 for x in range(0, 20)]
            # B = [x % 0x100 for x in range(10, 40)]
            # C = [x % 0x100 for x in range(5, 13)]
            D = [8] * 16

            for E in [A, B, C, D]:
                write_bytes(buf, E)
                buf.new()

            for E in [D, C, B, A]:
                assert buf.seek_frame(-1)
                tests = [(x, buf.read()) for x in E]
                assert all(list(map(lambda x: x[0] == x[1], tests)))

            order = [A, B, C, D]
            # DeltaFixedAllocBuffers doesn't repeat the first section
            if isinstance(buf, DeltaFixedAllocBuffers):
                order.pop(0)
            for E in order:
                assert buf.seek_frame(1)
                tests = [(x, buf.read()) for x in E]
                assert all(list(map(lambda x: x[0] == x[1], tests)))

    def test_delta_seek(self):
        buf = DeltaFixedAllocBuffers()
        A = [1] * 16
        B = [2] * 16
        C = [3] * 16

        write_bytes(buf, A)
        buf.new()
        write_bytes(buf, B)
        buf.new()
        write_bytes(buf, C)
        buf.new()

        assert buf.seek_frame(-1)
        tests = [(x, buf.read()) for x in C]
        assert all(list(map(lambda x: x[0] == x[1], tests)))

        assert buf.seek_frame(-1)
        tests = [(x, buf.read()) for x in B]
        assert all(list(map(lambda x: x[0] == x[1], tests)))

        assert buf.seek_frame(-1)
        tests = [(x, buf.read()) for x in A]
        assert all(list(map(lambda x: x[0] == x[1], tests)))

        # Hit buffer boundary
        assert not buf.seek_frame(-1)

        assert buf.seek_frame(1)
        tests = [(x, buf.read()) for x in B]
        assert all(list(map(lambda x: x[0] == x[1], tests)))

        assert buf.seek_frame(1)
        tests = [(x, buf.read()) for x in C]
        assert all(list(map(lambda x: x[0] == x[1], tests)))

        assert not buf.seek_frame(1)

    def test_compressed_buffer(self):
        buf = CompressedFixedAllocBuffers()

        write_bytes(buf, [0 for _ in range(10)])
        # Zeros are not written before a flush
        assert all(map(lambda x: x == FILL_VALUE, buf.buffer[:12]))

        # Now, we should see one pair of '0' and length
        buf.flush()
        assert all(map(lambda x: x == FILL_VALUE, buf.buffer[2:12]))
        assert buf.buffer[0] == 0
        assert buf.buffer[1] == 10

        # Second flush should do nothing
        buf.flush()
        assert all(map(lambda x: x == FILL_VALUE, buf.buffer[2:12]))
        assert buf.buffer[0] == 0
        assert buf.buffer[1] == 10

        # Overflow 8-bit counter and see two pairs
        write_bytes(buf, [0 for _ in range(256)])
        buf.flush()
        assert all(map(lambda x: x[0] == x[1], zip(buf.buffer[2:8], [0, 255, 0, 1] + [FILL_VALUE] * 2)))

        # Fit exactly within 8-bit counter
        write_bytes(buf, [0 for _ in range(255)])
        buf.flush()
        assert all(map(lambda x: x[0] == x[1], zip(buf.buffer[6:10], [0, 255] + [FILL_VALUE] * 4)))

    def test_delta_buffer(self):
        buf = DeltaFixedAllocBuffers()
        assert all(map(lambda x: x == FILL_VALUE, buf.buffer[:60]))
        assert all(map(lambda x: x == 0, buf.internal_buffer[:60]))

        write_bytes(buf, range(20))
        # The compression adds a [0, 1] prefix, as there is a zero, repeated once
        assert all(map(lambda x: x[0] == x[1], zip(buf.buffer[:60], [0, 1] + list(range(1, 20)) + [FILL_VALUE] * 40)))
        # The compression doesn't exist in the internal buffer
        assert all(map(lambda x: x[0] == x[1], zip(buf.internal_buffer[:60], list(range(20)) + [0] * 40)))
        buf.new()

        # Write same range(20), but add 0x80 to set the 7th bit.
        # This should result in a delta with 20 values of 128 as only
        # the 7th bit is the xor-difference from the frame above.
        write_bytes(buf, range(0x80, 0x80 + 20))
        assert all(
            map(
                lambda x: x[0] == x[1],
                zip(buf.buffer[:60], [0, 1] + list(range(1, 20)) + [0x80] * 20 + [FILL_VALUE] * 20)
            )
        )
        assert all(map(lambda x: x[0] == x[1], zip(buf.internal_buffer[:60], list(range(0x80, 0x80 + 20)) + [0] * 40)))
        buf.new()

        write_bytes(buf, [0xFF] * 20)
        assert all(
            map(
                lambda x: x[0] == x[1],
                zip(
                    buf.buffer[:61],
                    [0, 1] + list(range(1, 20)) + [0x80] * 20 + [x ^ 0xFF for x in list(range(0x80, 0x80 + 20))]
                )
            )
        )
        assert all(map(lambda x: x[0] == x[1], zip(buf.internal_buffer[:60], [0xFF] * 20 + [0] * 40)))
        buf.new()

    def test_delta_buffer_repeat_pattern(self):
        buf = DeltaFixedAllocBuffers()
        assert all(map(lambda x: x == FILL_VALUE, buf.buffer[:60]))
        assert all(map(lambda x: x == 0, buf.internal_buffer[:60]))

        # Initial frame will just show up directly in the underlying buffer
        write_bytes(buf, [0xAA] * 20)
        assert all(map(lambda x: x[0] == x[1], zip(buf.buffer[:60], [0xAA] * 20 + [FILL_VALUE] * 40)))
        assert all(map(lambda x: x[0] == x[1], zip(buf.internal_buffer[:60], [0xAA] * 20 + [0] * 40)))
        buf.new()

        write_bytes(buf, [0xAA] * 20)
        # The written data should be zeros and only get written on a call to new (flush)
        assert all(map(lambda x: x[0] == x[1], zip(buf.buffer[:60], [0xAA] * 20 + [FILL_VALUE] * 40)))
        assert all(map(lambda x: x[0] == x[1], zip(buf.internal_buffer[:60], [0xAA] * 20 + [0] * 40)))
        buf.new()
        # Test that we see a zero-prefix with a count of 20
        assert all(map(lambda x: x[0] == x[1], zip(buf.buffer[:60], [0xAA] * 20 + [0, 20] + [FILL_VALUE] * 38)))
        # The internal buffer always reflect the current image, so nothing has changed.
        assert all(map(lambda x: x[0] == x[1], zip(buf.internal_buffer[:60], [0xAA] * 20 + [0] * 40)))

        write_bytes(buf, [0xAA] * 20)
        buf.new()
        # Same as above, with an additional zero-prefix
        assert all(map(lambda x: x[0] == x[1], zip(buf.buffer[:60], [0xAA] * 20 + [0, 20, 0, 20] + [FILL_VALUE] * 36)))
        assert all(map(lambda x: x[0] == x[1], zip(buf.internal_buffer[:60], [0xAA] * 20 + [0] * 40)))

    def test_buffer_overrun(self):
        buf = FixedAllocBuffers()
        # Fill almost entire buffer, as we want this section to be removed when we write more than the buffer can
        # contain.
        write_bytes(buf, [0xAA] * (FIXED_BUFFER_SIZE-10))
        buf.new()
        # We should have the first [0] section plus the one above.
        assert len(buf.sections) == 2
        # Writing 20 bytes, should overrun the buffer, and remove the section above to make room.
        write_bytes(buf, [0xAA] * 20)
        # We can verify this already, that one section is available.
        assert len(buf.sections) == 1
        # When call .new(), we are back to 2 sections.
        buf.new()
        assert len(buf.sections) == 2



#
# License: See LICENSE.md file
# GitHub: https://github.com/Baekalfen/PyBoy
#

import base64
import hashlib
import io
import json
import os
import sys
import time
import zlib
from unittest import mock

import numpy as np
import pytest

from pyboy import PyBoy, WindowEvent, utils

event_filter = [
    WindowEvent.PRESS_SPEED_UP,
    WindowEvent.RELEASE_SPEED_UP,
    WindowEvent.SCREEN_RECORDING_TOGGLE,
    WindowEvent._INTERNAL_RENDERER_FLUSH,
]
# Set to true to reset tests
RESET_REPLAYS = False


def verify_screen_image_np(pyboy, saved_array):
    match = np.all(
        np.frombuffer(saved_array, dtype=np.uint8).reshape(144, 160, 3) ==
        pyboy.botsupport_manager().screen().screen_ndarray()
    )
    if not match and not os.environ.get("TEST_CI"):
        from PIL import Image
        original = Image.frombytes("RGB", (160, 144), np.frombuffer(saved_array, dtype=np.uint8).reshape(144, 160, 3))
        original.show()
        new = pyboy.botsupport_manager().screen().screen_image()
        new.show()
        import PIL.ImageChops
        PIL.ImageChops.difference(original, new).show()

    assert match


def verify_file_hash(rom_file, b64_target_hash):
    with open(rom_file, "rb") as f:
        m = hashlib.sha256()
        m.update(f.read())
        b64_romhash = base64.b64encode(m.digest()).decode("utf8")
        assert b64_romhash == b64_target_hash


def move_gif(game, dest):
    record_dir = "recordings"
    for _ in range(10):
        try:
            gif = sorted(filter(lambda x: game in x, os.listdir(record_dir)))[-1]
            os.replace(record_dir + "/" + gif, dest)
            break
        except:
            time.sleep(1)
    else:
        raise FileNotFoundError(f"Couldn't find gif to move for game {game}")


def replay(
    ROM,
    replay,
    window="headless",
    verify=False,
    record_gif=None,
    gif_destination=None,
    rewind=False,
    bootrom_file=None,
    overwrite=RESET_REPLAYS,
    gif_hash=None,
    randomize=False,
    padding_frames=0,
    stop_frame=-1,
    cgb=None,
):
    with open(replay, "rb") as f:
        recorded_input, b64_romhash, b64_state = json.loads(zlib.decompress(f.read()).decode("ascii"))

    verify_file_hash(ROM, b64_romhash)
    state_data = io.BytesIO(base64.b64decode(b64_state.encode("utf8"))) if b64_state is not None else None

    pyboy = PyBoy(
        ROM,
        window_type=window,
        bootrom_file=bootrom_file,
        disable_input=True,
        rewind=rewind,
        randomize=randomize,
        cgb=cgb,
        record_input=(RESET_REPLAYS and window in ["SDL2", "headless", "OpenGL"]),
    )
    pyboy.set_emulation_speed(0)
    if state_data is not None:
        pyboy.load_state(state_data)
    else:
        for _ in range(padding_frames):
            pyboy.tick()

    # Filters out the blacklisted events
    recorded_input = list(
        map(
            lambda event_tuple:
            (event_tuple[0], list(filter(lambda x: x not in event_filter, event_tuple[1])), event_tuple[2]),
            recorded_input
        )
    )

    frame_count = 0
    next_event = recorded_input.pop(0)

    recording = False
    while recorded_input != [] and stop_frame != frame_count:
        if record_gif is not None and (frame_count in record_gif):
            pyboy.send_input(WindowEvent.SCREEN_RECORDING_TOGGLE)
            recording ^= True

        if next_event[0] == frame_count:
            for e in next_event[1]:
                pyboy.send_input(e)

                if verify and not overwrite and frame_count > 1: # First frame or two might be wrong on old statefiles
                    verify_screen_image_np(pyboy, base64.b64decode(next_event[2].encode("utf8")))
            next_event = recorded_input.pop(0)
        frame_count += 1
        # if frame_count % 30 == 0:
        #     print(frame_count)
        #     breakpoint()
        pyboy.tick()

    print(frame_count)
    # If end-frame in record_gif is high than frame counter
    if recording:
        pyboy.send_input(WindowEvent.SCREEN_RECORDING_TOGGLE)
        # We need to run an extra cycle for the screen recording to save
        pyboy.tick()
        print(frame_count)
        recording ^= True

    if gif_destination:
        move_gif(pyboy.cartridge_title(), gif_destination)
        if gif_hash is not None and not overwrite and sys.platform == "darwin":
            verify_file_hash(gif_destination, gif_hash)

    if overwrite:
        with open(replay, "wb") as f:
            f.write(
                zlib.compress(
                    json.dumps((pyboy.plugin_manager.record_replay.recorded_input, b64_romhash, b64_state)).encode()
                )
            )

    pyboy.stop(save=False)


def test_pokemon(pokemon_blue_rom, boot_rom):
    replay(
        pokemon_blue_rom,
        "tests/replays/pokemon_blue.replay",
        stop_frame=1074,
        bootrom_file=boot_rom,
    )


def test_pokemon_gif1(pokemon_gold_rom, boot_rom):
    replay(
        pokemon_gold_rom,
        "tests/replays/pokemon_gold_gif.replay",
        record_gif=(1, 2714),
        gif_destination="extras/README/1.gif",
        bootrom_file=boot_rom,
    )


def test_pokemon_gif2(pokemon_blue_rom, boot_rom):
    replay(
        pokemon_blue_rom,
        "tests/replays/pokemon_blue_gif2.replay",
        record_gif=(0, 180),
        gif_destination="extras/README/2.gif",
        bootrom_file=boot_rom,
    )


def test_tetris(tetris_rom, boot_rom):
    replay(
        tetris_rom,
        "tests/replays/tetris.replay",
        bootrom_file=boot_rom,
    )


def test_supermarioland(supermarioland_rom, boot_rom):
    replay(
        supermarioland_rom,
        "tests/replays/supermarioland.replay",
        bootrom_file=boot_rom,
    )


def test_kirby(kirby_rom, boot_rom):
    replay(
        kirby_rom,
        "tests/replays/kirby_gif.replay",
        record_gif=(0, 360),
        gif_destination="extras/README/4.gif",
        bootrom_file=boot_rom,
    )


def test_rewind(supermarioland_rom, boot_rom):
    replay(
        supermarioland_rom,
        "tests/replays/supermarioland_rewind.replay",
        record_gif=(130, 544),
        gif_destination="extras/README/5.gif",
        rewind=True,
        bootrom_file=None,
    )



#
# License: See LICENSE.md file
# GitHub: https://github.com/Baekalfen/PyBoy
#
"""
Memory constants used internally to calculate tile and tile map addresses.
"""

VRAM_OFFSET = 0x8000
LCDC_OFFSET = 0xFF40
OAM_OFFSET = 0xFE00
LOW_TILEMAP = 0x1800 + VRAM_OFFSET
HIGH_TILEMAP = 0x1C00 + VRAM_OFFSET
LOW_TILEDATA = VRAM_OFFSET
LOW_TILEDATA_NTILES = 0x100
HIGH_TILEDATA = 0x800 + VRAM_OFFSET
TILES = 384
SPRITES = 40
ROWS, COLS = 144, 160



#
# License: See LICENSE.md file
# GitHub: https://github.com/Baekalfen/PyBoy
#

from pyboy.core.mb cimport Motherboard



cdef class TileMap:
    cdef Motherboard mb
    cdef bint signed_tile_data
    cdef bint _use_tile_objects
    cdef int map_offset
    cdef str _select
    cdef public tuple shape



#
# License: See LICENSE.md file
# GitHub: https://github.com/Baekalfen/PyBoy
#
"""
This class gives access to the frame buffer and other screen parameters of PyBoy.
"""

import numpy as np

from pyboy import utils
from pyboy.logging import get_logger

from .constants import COLS, ROWS

logger = get_logger(__name__)

try:
    from PIL import Image
except ImportError:
    Image = None


class Screen:
    """
    As part of the emulation, we generate a screen buffer in 32-bit RGBA format. This class has several helper methods
    to make it possible to read this buffer out.

    If you're making an AI or bot, it's highly recommended to _not_ use this class for detecting objects on the screen.
    It's much more efficient to use `pyboy.botsupport.BotSupportManager.tilemap_background`, `pyboy.botsupport.BotSupportManager.tilemap_window`, and
    `pyboy.botsupport.BotSupportManager.sprite` instead.
    """
    def __init__(self, mb):
        self.mb = mb

    def tilemap_position(self):
        """
        These coordinates define the offset in the tile map from where the top-left corner of the screen is place. Note
        that the tile map defines 256x256 pixels, but the screen can only show 160x144 pixels. When the offset is closer
        to the right or bottom edge than 160x144 pixels, the screen will wrap around and render from the opposite site
        of the tile map.

        For more details, see "7.4 Viewport" in the [report](https://github.com/Baekalfen/PyBoy/raw/master/extras/PyBoy.pdf),
        or the Pan Docs under [LCD Position and Scrolling](http://bgb.bircd.org/pandocs.htm#lcdpositionandscrolling).

        Returns
        -------
        tuple:
            Returns the tuple of registers ((SCX, SCY), (WX - 7, WY))
        """
        return (self.mb.lcd.getviewport(), self.mb.lcd.getwindowpos())

    def tilemap_position_list(self):
        """
        This function provides the screen (SCX, SCY) and window (WX. WY) position for each horizontal line in the
        screen buffer. These parameters are often used for visual effects, and some games will reset the registers at
        the end of each call to `pyboy.PyBoy.tick()`. For such games, `Screen.tilemap_position` becomes useless.

        See `Screen.tilemap_position` for more information.

        Returns
        -------
        list:
            Nested list of SCX, SCY, WX and WY for each scanline (144x4). Returns (0, 0, 0, 0) when LCD is off.
        """
        if self.mb.lcd._LCDC.lcd_enable:
            return [[line[0], line[1], line[2], line[3]] for line in self.mb.lcd.renderer._scanlineparameters]
        else:
            return [[0, 0, 0, 0] for line in range(144)]

    def raw_screen_buffer(self):
        """
        Provides a raw, unfiltered `bytes` object with the data from the screen. Check
        `Screen.raw_screen_buffer_format` to see which dataformat is used. The returned type and dataformat are
        subject to change.

        Use this, only if you need to bypass the overhead of `Screen.screen_image` or `Screen.screen_ndarray`.

        Returns
        -------
        bytes:
            92160 bytes of screen data in a `bytes` object.
        """
        return self.mb.lcd.renderer._screenbuffer_raw.tobytes()

    def raw_screen_buffer_dims(self):
        """
        Returns the dimensions of the raw screen buffer.

        Returns
        -------
        tuple:
            A two-tuple of the buffer dimensions. E.g. (160, 144).
        """
        return self.mb.lcd.renderer.buffer_dims

    def raw_screen_buffer_format(self):
        """
        Returns the color format of the raw screen buffer.

        Returns
        -------
        str:
            Color format of the raw screen buffer. E.g. 'RGB'.
        """
        return self.mb.lcd.renderer.color_format

    def screen_ndarray(self):
        """
        Provides the screen data in NumPy format. The dataformat is always RGB.

        Returns
        -------
        numpy.ndarray:
            Screendata in `ndarray` of bytes with shape (160, 144, 3)
        """
        return np.frombuffer(self.mb.lcd.renderer._screenbuffer_raw, dtype=np.uint8).reshape(ROWS, COLS, 4)[:, :, 1:]

    def screen_image(self):
        """
        Generates a PIL Image from the screen buffer.

        Convenient for screen captures, but might be a bottleneck, if you use it to train a neural network. In which
        case, read up on the `pyboy.botsupport` features, [Pan Docs](http://bgb.bircd.org/pandocs.htm) on tiles/sprites,
        and join our Discord channel for more help.

        Returns
        -------
        PIL.Image:
            RGB image of (160, 144) pixels
        """
        if not Image:
            logger.error("Cannot generate screen image. Missing dependency \"Pillow\".")
            return None

        # NOTE: Might have room for performance improvement
        # It's not possible to use the following, as the byte-order (endianess) isn't supported in Pillow
        # Image.frombytes('RGBA', self.buffer_dims, self.screen_buffer()).show()
        # FIXME: FORMAT IS BGR NOT RGB!!!
        return Image.fromarray(self.screen_ndarray()[:, :, [2, 1, 0]], "RGB")



#
# License: See LICENSE.md file
# GitHub: https://github.com/Baekalfen/PyBoy
#
import re

# Plugins and priority!
# E.g. DisableInput first
windows = ["WindowSDL2", "WindowOpenGL", "WindowHeadless", "WindowDummy", "Debug"]
game_wrappers = [
    "GameWrapperSuperMarioLand", "GameWrapperTetris", "GameWrapperKirbyDreamLand", "GameWrapperPokemonGen1"
]
plugins = [
    "DisableInput", "AutoPause", "RecordReplay", "Rewind", "ScreenRecorder", "ScreenshotRecorder"
] + game_wrappers
all_plugins = windows + plugins


def to_snake_case(s):
    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", s)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()


def skip_lines(iterator, stop):
    # Skip old lines
    while True:
        if next(line_iter).strip().startswith(stop):
            break


if __name__ == "__main__":
    out_lines = []
    with open("manager.py", "r") as f:
        line_iter = iter(f.readlines())
        while True:
            line = next(line_iter, None)
            if line is None:
                break

            # Find place to inject
            if line.strip().startswith("# foreach"):

                lines = [line.strip() + "\n"]
                indentation = " " * line.index("# foreach")

                skip_lines(line_iter, "# foreach end")

                _, foreach, plugin_type, fun = line.strip().split(" ", 3)
                for p in eval(plugin_type):
                    p_name = to_snake_case(p)
                    lines.append(f"if self.{p_name}_enabled:\n")
                    # lines.append(f"    {var_name} = self.{p_name}\n")
                    for sub_fun in fun.split(", "):
                        sub_fun = sub_fun.replace("[]", f"self.{p_name}")
                        lines.append(f"    {sub_fun}\n")

                lines.append("# foreach end\n")
                out_lines.extend([indentation + l for l in lines])
            elif line.strip().startswith("# plugins_enabled"):

                lines = [line.strip() + "\n"]
                indentation = " " * line.index("# plugins_enabled")

                skip_lines(line_iter, "# plugins_enabled end")

                for p in all_plugins:
                    p_name = to_snake_case(p)
                    lines.append(f"self.{p_name} = {p}(pyboy, mb, pyboy_argv)\n")
                    lines.append(f"self.{p_name}_enabled = self.{p_name}.enabled()\n")

                lines.append("# plugins_enabled end\n")
                out_lines.extend([indentation + l for l in lines])
            elif line.strip().startswith("# yield_plugins"):

                lines = [line.strip() + "\n"]
                indentation = " " * line.index("# yield_plugins")

                skip_lines(line_iter, "# yield_plugins end")

                for p in all_plugins:
                    p_name = to_snake_case(p)
                    lines.append(f"yield {p}.argv\n")

                lines.append("# yield_plugins end\n")
                out_lines.extend([indentation + l for l in lines])
            elif line.strip().startswith("# imports"):

                lines = [line.strip() + "\n"]
                indentation = " " * line.index("# imports")

                skip_lines(line_iter, "# imports end")

                for p in all_plugins:
                    p_name = to_snake_case(p)
                    lines.append(f"from pyboy.plugins.{p_name} import {p} # isort:skip\n")

                lines.append("# imports end\n")
                out_lines.extend([indentation + l for l in lines])
            elif line.strip().startswith("# gamewrapper"):

                lines = [line.strip() + "\n"]
                indentation = " " * line.index("# gamewrapper")

                skip_lines(line_iter, "# gamewrapper end")

                for p in game_wrappers:
                    p_name = to_snake_case(p)
                    lines.append(f"if self.{p_name}_enabled: return self.{p_name}\n")

                lines.append("# gamewrapper end\n")
                out_lines.extend([indentation + l for l in lines])
            else:
                out_lines.append(line)

    with open("manager.py", "w") as f:
        f.writelines(out_lines)

    out_lines = []
    with open("manager.pxd", "r") as f:
        line_iter = iter(f.readlines())
        while True:
            line = next(line_iter, None)
            if line is None:
                break

            # Find place to inject
            if line.strip().startswith("# plugin_cdef"):

                lines = [line.strip() + "\n"]
                indentation = " " * line.index("# plugin_cdef")

                skip_lines(line_iter, "# plugin_cdef end")

                for p in all_plugins:
                    p_name = to_snake_case(p)
                    lines.append(f"cdef public {p} {p_name}\n")

                for p in all_plugins:
                    p_name = to_snake_case(p)
                    lines.append(f"cdef bint {p_name}_enabled\n")

                lines.append("# plugin_cdef end\n")
                out_lines.extend([indentation + l for l in lines])
            elif line.strip().startswith("# imports"):

                lines = [line.strip() + "\n"]
                indentation = " " * line.index("# imports")

                skip_lines(line_iter, "# imports end")

                for p in all_plugins:
                    p_name = to_snake_case(p)
                    lines.append(f"from pyboy.plugins.{p_name} cimport {p}\n")

                lines.append("# imports end\n")
                out_lines.extend([indentation + l for l in lines])
            else:
                out_lines.append(line)

    with open("manager.pxd", "w") as f:
        f.writelines(out_lines)

    out_lines = []
    with open("__init__.py", "r") as f:
        line_iter = iter(f.readlines())
        while True:
            line = next(line_iter, None)
            if line is None:
                break

            # Find place to inject
            if line.strip().startswith("# docs exclude"):

                lines = [line.strip() + "\n"]
                indentation = " " * line.index("# docs exclude")

                skip_lines(line_iter, "# docs exclude end")

                for p in (set(all_plugins) - set(game_wrappers)) | set(["manager", "manager_gen"]):
                    p_name = to_snake_case(p)
                    lines.append(f"\"{p_name}\": False,\n")

                lines.append("# docs exclude end\n")
                out_lines.extend([indentation + l for l in lines])
            else:
                out_lines.append(line)

    with open("__init__.py", "w") as f:
        f.writelines(out_lines)



#
# License: See LICENSE.md file
# GitHub: https://github.com/Baekalfen/PyBoy
#

from pyboy.logging.logging cimport Logger
from pyboy.plugins.base_plugin cimport PyBoyPlugin


cdef Logger logger

cdef class DisableInput(PyBoyPlugin):
    pass





#
# License: See LICENSE.md file
# GitHub: https://github.com/Baekalfen/PyBoy
#

# from .base_mbc import BaseMBC, ROMOnly
from .cartridge import load_cartridge

# from .mbc1 import MBC1
# from .mbc2 import MBC2
# from .mbc3 import MBC3
# from .mbc5 import MBC5



#
# License: See LICENSE.md file
# GitHub: https://github.com/Baekalfen/PyBoy
#

import os
import re
import zlib
from array import array
from base64 import b64decode
from ctypes import c_void_p

from pyboy import utils
from pyboy.botsupport import constants, tilemap
from pyboy.botsupport.sprite import Sprite
from pyboy.plugins.base_plugin import PyBoyWindowPlugin
from pyboy.plugins.window_sdl2 import sdl2_event_pump
from pyboy.utils import WindowEvent

try:
    from cython import compiled
    cythonmode = compiled
except ImportError:
    cythonmode = False

try:
    import sdl2
except ImportError:
    sdl2 = None

import pyboy

logger = pyboy.logging.get_logger(__name__)

# Mask colors:
COLOR = 0x00000000
# MASK = 0x00C0C000
COLOR_BACKGROUND = 0x00C0C000
COLOR_WINDOW = 0xC179D400

# Additive colors
HOVER = 0xFF0000
mark_counter = 0
# By using a set, we avoid duplicates
marked_tiles = set([])
MARK = array("I", [0xFF000000, 0xFFC00000, 0xFFFC0000, 0x00FFFF00, 0xFF00FF00])

SPRITE_BACKGROUND = COLOR_BACKGROUND


class MarkedTile:
    def __init__(
        self,
        event=WindowEvent._INTERNAL_MARK_TILE,
        tile_identifier=-1,
        mark_id="",
        mark_color=0,
        sprite_height=8,
        sprite=False
    ):
        self.tile_identifier = tile_identifier
        self.mark_id = mark_id
        self.mark_color = mark_color
        self.sprite_height = sprite_height
        if mark_id == "TILE":
            # TODO: Use __str__ of the Tile and Sprite classes
            logger.info(f"Marked Tile - identifier: {tile_identifier}")
        elif mark_id == "SPRITE":
            logger.info(f"Marked Sprite - tile identifier: {tile_identifier}, sprite height: {sprite_height}")
        else:
            logger.info(f"Marked {mark_id} - tile identifier: {tile_identifier}")

    def __hash__(self):
        return hash(self.tile_identifier)


class Debug(PyBoyWindowPlugin):
    argv = [("-d", "--debug", {
        "action": "store_true",
        "help": "Enable emulator debugging mode"
    }), ("--breakpoints", {
        "type": str,
        "help": "Add breakpoints on start-up (internal use)"
    })]

    def __init__(self, pyboy, mb, pyboy_argv):
        super().__init__(pyboy, mb, pyboy_argv)

        if not self.enabled():
            return

        self.cgb = mb.cgb

        self.rom_symbols = {}
        if pyboy_argv.get("ROM"):
            gamerom_file_no_ext, rom_ext = os.path.splitext(pyboy_argv.get("ROM"))
            for sym_ext in [".sym", rom_ext + ".sym"]:
                sym_path = gamerom_file_no_ext + sym_ext
                if os.path.isfile(sym_path):
                    with open(sym_path) as f:
                        for _line in f.readlines():
                            line = _line.strip()
                            if line == "":
                                continue
                            elif line.startswith(";"):
                                continue
                            elif line.startswith("["):
                                # Start of key group
                                # [labels]
                                # [definitions]
                                continue

                            try:
                                bank, addr, sym_label = re.split(":| ", line.strip())
                                bank = int(bank, 16)
                                addr = int(addr, 16)
                                if not bank in self.rom_symbols:
                                    self.rom_symbols[bank] = {}

                                self.rom_symbols[bank][addr] = sym_label
                            except ValueError as ex:
                                logger.warning("Skipping .sym line: %s", line.strip())

        self.sdl2_event_pump = self.pyboy_argv.get("window_type") != "SDL2"
        if self.sdl2_event_pump:
            sdl2.SDL_Init(sdl2.SDL_INIT_VIDEO)

        # self.scale = 2
        window_pos = 0

        self.tile1 = TileViewWindow(
            pyboy,
            mb,
            pyboy_argv,
            scale=2,
            title="Background",
            width=256,
            height=256,
            pos_x=0,
            pos_y=0,
            window_map=False,
            scanline_x=0,
            scanline_y=1
        )
        window_pos += (256 * self.tile1.scale)

        self.tile2 = TileViewWindow(
            pyboy,
            mb,
            pyboy_argv,
            scale=2,
            title="Window",
            width=256,
            height=256,
            pos_x=window_pos,
            pos_y=0,
            window_map=True,
            scanline_x=2,
            scanline_y=3
        )
        window_pos += (256 * self.tile2.scale)

        self.spriteview = SpriteViewWindow(
            pyboy,
            mb,
            pyboy_argv,
            scale=2,
            title="Sprite View",
            width=constants.COLS,
            height=constants.ROWS,
            pos_x=window_pos,
            pos_y=0
        )

        self.sprite = SpriteWindow(
            pyboy,
            mb,
            pyboy_argv,
            scale=3,
            title="Sprite Data",
            width=8 * 10,
            height=16 * 4,
            pos_x=window_pos,
            pos_y=self.spriteview.height * 2 + 68
        )
        window_pos += (constants.COLS * self.spriteview.scale)

        self.memory = MemoryWindow(
            pyboy, mb, pyboy_argv, scale=1, title="Memory", width=8 * 60, height=16 * 36, pos_x=window_pos, pos_y=0
        )
        window_pos += 8 * 60

        window_pos = 0
        tile_data_width = 16 * 8 # Change the 16 to however wide you want the tile window
        tile_data_height = ((constants.TILES * 8) // tile_data_width) * 8
        self.tiledata0 = TileDataWindow(
            pyboy,
            mb,
            pyboy_argv,
            scale=3,
            title="Tile Data0",
            width=tile_data_width,
            height=tile_data_height,
            pos_x=window_pos,
            pos_y=(256 * self.tile1.scale) + 128
        )
        if self.cgb:
            window_pos += 512
            self.tiledata1 = TileDataWindow(
                pyboy,
                mb,
                pyboy_argv,
                scale=3,
                title="Tile Data1",
                width=tile_data_width,
                height=tile_data_height,
                pos_x=window_pos,
                pos_y=(256 * self.tile1.scale) + 128
            )

        for _b in (self.pyboy_argv.get("breakpoints") or "").split(","):
            b = _b.strip()
            if b == "":
                continue

            bank_addr = self.parse_bank_addr_sym_label(b)
            if bank_addr is None:
                logger.error("Couldn't parse address or label: %s", b)
                pass
            else:
                self.mb.add_breakpoint(*bank_addr)
                logger.info("Added breakpoint for address or label: %s", b)

    def post_tick(self):
        self.tile1.post_tick()
        self.tile2.post_tick()
        self.tiledata0.post_tick()
        if self.cgb:
            self.tiledata1.post_tick()
        self.sprite.post_tick()
        self.spriteview.post_tick()
        self.memory.post_tick()

    def handle_events(self, events):
        if self.sdl2_event_pump:
            events = sdl2_event_pump(events)
        events = self.tile1.handle_events(events)
        events = self.tile2.handle_events(events)
        events = self.tiledata0.handle_events(events)
        if self.cgb:
            events = self.tiledata1.handle_events(events)
        events = self.sprite.handle_events(events)
        events = self.spriteview.handle_events(events)
        events = self.memory.handle_events(events)
        return events

    def stop(self):
        if self.sdl2_event_pump:
            sdl2.SDL_Quit()

    def enabled(self):
        if self.pyboy_argv.get("debug"):
            if not sdl2:
                logger.error("Failed to import sdl2, needed for debug window")
                return False
            else:
                return True
        else:
            return False

    def parse_bank_addr_sym_label(self, command):
        if ":" in command:
            bank, addr = command.split(":")
            bank = int(bank, 16)
            addr = int(addr, 16)
            return bank, addr
        else:
            for bank, addresses in self.rom_symbols.items():
                for addr, label in addresses.items():
                    if label == command:
                        return bank, addr
        return None

    def handle_breakpoint(self):
        while True:
            self.post_tick()

            if self.mb.cpu.PC < 0x4000:
                bank = 0
            else:
                bank = self.mb.cartridge.rombank_selected
            sym_label = self.rom_symbols.get(bank, {}).get(self.mb.cpu.PC, "")

            print(self.mb.cpu.dump_state(sym_label))
            cmd = input()

            if cmd == "c" or cmd.startswith("c "):
                # continue
                if cmd.startswith("c "):
                    _, command = cmd.split(" ", 1)
                    bank_addr = self.parse_bank_addr_sym_label(command)
                    if bank_addr is None:
                        print("Couldn't parse address or label!")
                    else:
                        # TODO: Possibly add a counter of 1, and remove the breakpoint after hitting it the first time
                        self.mb.add_breakpoint(*bank_addr)
                        break
                else:
                    break
            elif cmd == "sl":
                for bank, addresses in self.rom_symbols.items():
                    for addr, label in addresses.items():
                        print(f"{bank:02X}:{addr:04X} {label}")
            elif cmd == "bl":
                for bank, addr in self.mb.breakpoints_list:
                    print(f"{bank:02X}:{addr:04X} {self.rom_symbols.get(bank, {}).get(addr, '')}")
            elif cmd == "b" or cmd.startswith("b "):
                if cmd.startswith("b "):
                    _, command = cmd.split(" ", 1)
                else:
                    command = input(
                        'Write address in the format of "00:0150" or search for a symbol label like "Main"\n'
                    )

                bank_addr = self.parse_bank_addr_sym_label(command)
                if bank_addr is None:
                    print("Couldn't parse address or label!")
                else:
                    self.mb.add_breakpoint(*bank_addr)

            elif cmd == "d":
                # Remove current breakpoint

                # TODO: Share this code with breakpoint_reached
                for i, (bank, pc) in enumerate(self.mb.breakpoints_list):
                    if self.mb.cpu.PC == pc and (
                        (pc < 0x4000 and bank == 0 and not self.mb.bootrom_enabled) or \
                        (0x4000 <= pc < 0x8000 and self.mb.cartridge.rombank_selected == bank) or \
                        (0xA000 <= pc < 0xC000 and self.mb.cartridge.rambank_selected == bank) or \
                        (pc < 0x100 and bank == -1 and self.mb.bootrom_enabled)
                    ):
                        break
                else:
                    print("Breakpoint couldn't be deleted for current PC. Not Found.")
                    continue
                print(f"Removing breakpoint: {bank}:{pc}")
                self.mb.remove_breakpoint(i)
            elif cmd == "pdb":
                # Start pdb
                import pdb
                pdb.set_trace()
                break
            else:
                # Step once
                self.mb.breakpoint_latch = 1
                self.mb.tick()


def make_buffer(w, h):
    buf = array("B", [0x55] * (w*h*4))
    buf0 = memoryview(buf).cast("I", shape=(h, w))
    buf_p = c_void_p(buf.buffer_info()[0])
    return buf, buf0, buf_p


class BaseDebugWindow(PyBoyWindowPlugin):
    def __init__(self, pyboy, mb, pyboy_argv, *, scale, title, width, height, pos_x, pos_y):
        super().__init__(pyboy, mb, pyboy_argv)
        self.cgb = mb.cgb
        self.scale = scale
        self.width, self.height = width, height
        self.base_title = title
        self.hover_x = -1
        self.hover_y = -1

        self._window = sdl2.SDL_CreateWindow(
            self.base_title.encode("utf8"), pos_x, pos_y, width * scale, height * scale, sdl2.SDL_WINDOW_RESIZABLE
        )
        self.window_id = sdl2.SDL_GetWindowID(self._window)

        self.buf, self.buf0, self.buf_p = make_buffer(width, height)

        self._sdlrenderer = sdl2.SDL_CreateRenderer(self._window, -1, sdl2.SDL_RENDERER_ACCELERATED)
        sdl2.SDL_RenderSetLogicalSize(self._sdlrenderer, width, height)
        self._sdltexturebuffer = sdl2.SDL_CreateTexture(
            self._sdlrenderer, sdl2.SDL_PIXELFORMAT_RGBA8888, sdl2.SDL_TEXTUREACCESS_STATIC, width, height
        )

    def handle_events(self, events):
        # Feed events into the loop
        for event in events:
            if event == WindowEvent._INTERNAL_MOUSE:
                if event.window_id == self.window_id:
                    self.hover_x = event.mouse_x
                    self.hover_y = event.mouse_y
                else:
                    self.hover_x = -1
                    self.hover_y = -1

        return events

    def stop(self):
        sdl2.SDL_DestroyWindow(self._window)

    def update_title(self):
        pass

    def post_tick(self):
        self.update_title()
        sdl2.SDL_UpdateTexture(self._sdltexturebuffer, None, self.buf_p, self.width * 4)
        sdl2.SDL_RenderCopy(self._sdlrenderer, self._sdltexturebuffer, None, None)
        sdl2.SDL_RenderPresent(self._sdlrenderer)
        sdl2.SDL_RenderClear(self._sdlrenderer)

    ##########################
    # Internal functions
    def copy_tile(self, from_buffer, t, xx, yy, to_buffer, hflip, vflip, palette):
        for y in range(8):
            _y = 7 - y if vflip else y
            for x in range(8):
                _x = 7 - x if hflip else x
                to_buffer[yy + y, xx + x] = palette[from_buffer[_y + t*8][_x]]

    def mark_tile(self, x, y, color, height, width, grid):
        tw = width # Tile width
        th = height # Tile height
        if grid:
            xx = x - (x%tw)
            yy = y - (y%th)
        else:
            xx = x
            yy = y
        for i in range(th):
            if 0 <= (yy + i) < self.height and 0 <= xx < self.width:
                self.buf0[yy + i][xx] = color
        for i in range(tw):
            if 0 <= (yy) < self.height and 0 <= xx + i < self.width:
                self.buf0[yy, xx + i] = color
        for i in range(tw):
            if 0 <= (yy + th - 1) < self.height and 0 <= xx + i < self.width:
                self.buf0[yy + th - 1][xx + i] = color
        for i in range(th):
            if 0 <= (yy + i) < self.height and 0 <= xx + tw - 1 < self.width:
                self.buf0[yy + i][xx + tw - 1] = color


class TileViewWindow(BaseDebugWindow):
    def __init__(self, *args, window_map, scanline_x, scanline_y, **kwargs):
        super().__init__(*args, **kwargs)
        self.scanline_x, self.scanline_y = scanline_x, scanline_y
        self.color = COLOR_WINDOW if window_map else COLOR_BACKGROUND

        if not cythonmode:
            self.tilemap = tilemap.TileMap(self.mb, "WINDOW" if window_map else "BACKGROUND")

    def __cinit__(self, pyboy, mb, *args, window_map, **kwargs):
        self.tilemap = tilemap.TileMap(self.mb, "WINDOW" if window_map else "BACKGROUND")

    def post_tick(self):
        # Updating screen buffer by copying tiles from cache
        mem_offset = self.tilemap.map_offset - constants.VRAM_OFFSET
        for n in range(mem_offset, mem_offset + 0x400):
            tile_index = self.mb.lcd.VRAM0[n]

            # Check the tile source and add offset
            # http://problemkaputt.de/pandocs.htm#lcdcontrolregister
            # BG & Window Tile Data Select   (0=8800-97FF, 1=8000-8FFF)
            if self.mb.lcd._LCDC.tiledata_select == 0:
                # (x ^ 0x80 - 128) to convert to signed, then add 256 for offset (reduces to + 128)
                tile_index = (tile_index ^ 0x80) + 128

            tile_column = (n-mem_offset) % 32
            tile_row = (n-mem_offset) // 32

            # tilecache = None
            # palette_rgb = None
            if self.cgb:
                palette, vbank, horiflip, vertflip, bg_priority = self.renderer._cgb_get_background_map_attributes(
                    self.mb.lcd, n
                )
                self.renderer.update_tilecache1(self.mb.lcd, tile_index, 1)
                self.tilecache = (self.renderer._tilecache1 if vbank else self.renderer._tilecache0)
                self.palette_rgb = self.mb.lcd.ocpd.palette_mem_rgb # TODO: Select palette by adding offset
            else:
                # Fake palette index
                palette = 0
                self.renderer.update_tilecache0(self.mb.lcd, tile_index, 0)
                self.tilecache = self.renderer._tilecache0
                horiflip, vertflip = False, False
                self.palette_rgb = self.mb.lcd.BGP.palette_mem_rgb

            self.copy_tile(
                self.tilecache, tile_index, tile_column * 8, tile_row * 8, self.buf0, horiflip, vertflip,
                self.palette_rgb
            )

        self.draw_overlay()
        BaseDebugWindow.post_tick(self)

    def handle_events(self, events):
        global mark_counter, marked_tiles

        self.tilemap.refresh_lcdc()

        # Feed events into the loop
        events = BaseDebugWindow.handle_events(self, events)
        for event in events:
            if event == WindowEvent._INTERNAL_MOUSE and event.window_id == self.window_id:
                if event.mouse_button == 0:
                    tile_x, tile_y = event.mouse_x // 8, event.mouse_y // 8
                    tile_identifier = self.tilemap.tile_identifier(tile_x, tile_y)
                    logger.info(f"Tile clicked on %d, %d", tile_x, tile_y)
                    marked_tiles.add(
                        MarkedTile(tile_identifier=tile_identifier, mark_id="TILE", mark_color=MARK[mark_counter])
                    )
                    mark_counter += 1
                    mark_counter %= len(MARK)
                elif event.mouse_button == 1:
                    marked_tiles.clear()
            elif event == WindowEvent._INTERNAL_MARK_TILE:
                marked_tiles.add(event.tile_identifier)

        return events

    def update_title(self):
        title = self.base_title
        title += " [HIGH MAP 0x9C00-0x9FFF]" if self.tilemap.map_offset == constants.HIGH_TILEMAP else \
            " [LOW MAP 0x9800-0x9BFF]"
        title += " [HIGH DATA (SIGNED) 0x8800-0x97FF]" if self.tilemap.signed_tile_data else \
            " [LOW DATA (UNSIGNED) 0x8000-0x8FFF]"
        if self.tilemap._select == "WINDOW":
            title += " [Window]"
        if self.tilemap._select == "BACKGROUND":
            title += " [Background]"
        sdl2.SDL_SetWindowTitle(self._window, title.encode("utf8"))

    def draw_overlay(self):
        global marked_tiles
        scanlineparameters = self.pyboy.botsupport_manager().screen().tilemap_position_list()

        background_view = self.scanline_x == 0

        # TODO: Refactor this
        # Mark screen area
        for y in range(constants.ROWS):
            xx = int(scanlineparameters[y][self.scanline_x])
            yy = int(scanlineparameters[y][self.scanline_y])

            if background_view: # Background
                # Wraps around edges of the screen
                if y == 0 or y == constants.ROWS - 1: # Draw top/bottom bar
                    for x in range(constants.COLS):
                        self.buf0[(yy+y) % 0xFF][(xx+x) % 0xFF] = COLOR
                else: # Draw body
                    self.buf0[(yy+y) % 0xFF][xx % 0xFF] = COLOR
                    for x in range(constants.COLS):
                        self.buf0[(yy+y) % 0xFF][(xx+x) % 0xFF] &= self.color
                    self.buf0[(yy+y) % 0xFF][(xx + constants.COLS) % 0xFF] = COLOR
            else: # Window
                # Takes a cut of the screen
                xx = -xx
                yy = -yy
                if yy + y == 0 or y == constants.ROWS - 1: # Draw top/bottom bar
                    for x in range(constants.COLS):
                        if 0 <= xx + x < constants.COLS:
                            self.buf0[yy + y, xx + x] = COLOR
                else: # Draw body
                    if 0 <= yy + y:
                        self.buf0[yy + y][max(xx, 0)] = COLOR
                        for x in range(constants.COLS):
                            if 0 <= xx + x < constants.COLS:
                                self.buf0[yy + y, xx + x] &= self.color
                        self.buf0[yy + y, xx + constants.COLS] = COLOR

        # Mark selected tiles
        for t, match in zip(
            marked_tiles, self.tilemap.search_for_identifiers([m.tile_identifier for m in marked_tiles])
        ):
            for row, column in match:
                self.mark_tile(column * 8, row * 8, t.mark_color, 8, 8, True)
        if self.hover_x != -1:
            self.mark_tile(self.hover_x, self.hover_y, HOVER, 8, 8, True)

        # Mark current scanline directly from LY,SCX,SCY,WX,WY
        if background_view:
            for x in range(constants.COLS):
                self.buf0[(self.mb.lcd.SCY + self.mb.lcd.LY) % 0xFF][(self.mb.lcd.SCX + x) % 0xFF] = 0xFF00CE12
        else:
            for x in range(constants.COLS):
                self.buf0[(self.mb.lcd.WY + self.mb.lcd.LY) % 0xFF][(self.mb.lcd.WX + x) % 0xFF] = 0xFF00CE12


class TileDataWindow(BaseDebugWindow):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tilecache_select = 0 if "0" in kwargs.get("title") else 1

    def post_tick(self):
        # TODO: We could select different palettes on CGB
        if self.tilecache_select:
            tilecache = self.renderer._tilecache1
        else:
            tilecache = self.renderer._tilecache0

        if self.cgb:
            self.palette_rgb = self.mb.lcd.bcpd.palette_mem_rgb # TODO: Select palette by adding offset
        else:
            self.palette_rgb = self.mb.lcd.BGP.palette_mem_rgb

        for t in range(constants.TILES):
            if self.tilecache_select:
                self.renderer.update_tilecache1(self.mb.lcd, t, 1)
            else:
                self.renderer.update_tilecache0(self.mb.lcd, t, 0)
            xx = (t*8) % self.width
            yy = ((t*8) // self.width) * 8
            self.copy_tile(tilecache, t, xx, yy, self.buf0, False, False, self.palette_rgb)

        self.draw_overlay()
        BaseDebugWindow.post_tick(self)

    def handle_events(self, events):
        global mark_counter, marked_tiles
        # Feed events into the loop
        events = BaseDebugWindow.handle_events(self, events)
        for event in events:
            if event == WindowEvent._INTERNAL_MOUSE and event.window_id == self.window_id:
                if event.mouse_button == 0:
                    tile_x, tile_y = event.mouse_x // 8, event.mouse_y // 8
                    tile_identifier = tile_y * (self.width // 8) + tile_x
                    marked_tiles.add(
                        MarkedTile(tile_identifier=tile_identifier, mark_id="TILE", mark_color=MARK[mark_counter])
                    )
                    mark_counter += 1
                    mark_counter %= len(MARK)
                elif event.mouse_button == 1:
                    marked_tiles.clear()
            elif event == WindowEvent._INTERNAL_MARK_TILE:
                marked_tiles.add(event.tile_identifier)
        return events

    def draw_overlay(self):
        # Mark selected tiles
        for t in marked_tiles:
            column = t.tile_identifier % (self.width // 8)
            row = t.tile_identifier // (self.width // 8)
            # Yes, we are using the height as width. This is because we present the tile data from left to right,
            # but the sprites with a height of 16, renders them stacked ontop of each other.
            self.mark_tile(column * 8, row * 8, t.mark_color, t.sprite_height, 8, True)


class SpriteWindow(BaseDebugWindow):
    def post_tick(self):

        # TODO: Could we use scanline_sprites with modified arguments to render all of this?
        sprite_height = 16 if self.mb.lcd._LCDC.sprite_height else 8
        for n in range(0, 0xA0, 4):
            # x = lcd.OAM[n]
            # y = lcd.OAM[n+1]
            t = self.mb.lcd.OAM[n + 2]
            attributes = self.mb.lcd.OAM[n + 3]
            xx = ((n//4) * 8) % self.width
            yy = (((n//4) * 8) // self.width) * sprite_height

            if self.cgb:
                palette = attributes & 0b111
                if attributes & 0b1000:
                    self.renderer.update_spritecache1(self.mb.lcd, t, 1)
                    if self.mb.lcd._LCDC.sprite_height:
                        self.renderer.update_spritecache1(self.mb.lcd, t + 1, 1)
                    self.spritecache = self.renderer._spritecache1
                else:
                    self.renderer.update_spritecache0(self.mb.lcd, t, 0)
                    if self.mb.lcd._LCDC.sprite_height:
                        self.renderer.update_spritecache0(self.mb.lcd, t + 1, 0)
                    self.spritecache = self.renderer._spritecache0
                self.palette_rgb = self.mb.lcd.ocpd.palette_mem_rgb # TODO: Select palette by adding offset
            else:
                # Fake palette index
                palette = 0
                if attributes & 0b10000:
                    self.renderer.update_spritecache1(self.mb.lcd, t, 0)
                    if self.mb.lcd._LCDC.sprite_height:
                        self.renderer.update_spritecache1(self.mb.lcd, t + 1, 0)
                    self.spritecache = self.renderer._spritecache1
                    self.palette_rgb = self.mb.lcd.OBP1.palette_mem_rgb
                else:
                    self.renderer.update_spritecache0(self.mb.lcd, t, 0)
                    if self.mb.lcd._LCDC.sprite_height:
                        self.renderer.update_spritecache0(self.mb.lcd, t + 1, 0)
                    self.spritecache = self.renderer._spritecache0
                    self.palette_rgb = self.mb.lcd.OBP0.palette_mem_rgb

            self.copy_tile(self.spritecache, t, xx, yy, self.buf0, False, False, self.palette_rgb)
            if sprite_height:
                self.copy_tile(self.spritecache, t + 1, xx, yy + 8, self.buf0, False, False, self.palette_rgb)

        self.draw_overlay()
        BaseDebugWindow.post_tick(self)

    def handle_events(self, events):
        global mark_counter, marked_tiles

        # Feed events into the loop
        events = BaseDebugWindow.handle_events(self, events)

        sprite_height = 16 if self.mb.lcd._LCDC.sprite_height else 8
        for event in events:
            if event == WindowEvent._INTERNAL_MOUSE and event.window_id == self.window_id:
                if event.mouse_button == 0:
                    tile_x, tile_y = event.mouse_x // 8, event.mouse_y // sprite_height
                    sprite_identifier = tile_y * (self.width // 8) + tile_x
                    if sprite_identifier > constants.SPRITES:
                        # Out of bounds
                        continue
                    sprite = Sprite(self.mb, sprite_identifier)
                    marked_tiles.add(
                        MarkedTile(
                            tile_identifier=sprite.tile_identifier,
                            mark_id="SPRITE",
                            mark_color=MARK[mark_counter],
                            sprite_height=sprite_height,
                            sprite=True,
                        )
                    )
                    mark_counter += 1
                    mark_counter %= len(MARK)
                elif event.mouse_button == 1:
                    marked_tiles.clear()
            elif event == WindowEvent._INTERNAL_MARK_TILE:
                marked_tiles.add(event.tile_identifier)
        return events

    def draw_overlay(self):
        sprite_height = 16 if self.mb.lcd._LCDC.sprite_height else 8
        # Mark selected tiles
        for m, matched_sprites in zip(
            marked_tiles,
            self.pyboy.botsupport_manager().sprite_by_tile_identifier([m.tile_identifier for m in marked_tiles])
        ):
            for sprite_index in matched_sprites:
                xx = (sprite_index*8) % self.width
                yy = ((sprite_index*8) // self.width) * sprite_height
                self.mark_tile(xx, yy, m.mark_color, sprite_height, 8, True)

        if self.hover_x != -1:
            self.mark_tile(self.hover_x, self.hover_y, HOVER, sprite_height, 8, True)

    def update_title(self):
        title = self.base_title
        title += " [8x16]" if self.mb.lcd._LCDC.sprite_height else " [8x8]"
        sdl2.SDL_SetWindowTitle(self._window, title.encode("utf8"))


class SpriteViewWindow(BaseDebugWindow):
    def post_tick(self):
        for y in range(constants.ROWS):
            for x in range(constants.COLS):
                self.buf0[y, x] = SPRITE_BACKGROUND

        for ly in range(144):
            self.mb.lcd.renderer.scanline_sprites(self.mb.lcd, ly, self.buf0, True)

        self.draw_overlay()
        BaseDebugWindow.post_tick(self)

    def draw_overlay(self):
        sprite_height = 16 if self.mb.lcd._LCDC.sprite_height else 8
        # Mark selected tiles
        for m, matched_sprites in zip(
            marked_tiles,
            self.pyboy.botsupport_manager().sprite_by_tile_identifier([m.tile_identifier for m in marked_tiles])
        ):
            for sprite_index in matched_sprites:
                sprite = Sprite(self.mb, sprite_index)
                self.mark_tile(sprite.x, sprite.y, m.mark_color, sprite_height, 8, False)

    def update_title(self):
        title = self.base_title
        title += " " if self.mb.lcd._LCDC.sprite_enable else " [Disabled]"
        sdl2.SDL_SetWindowTitle(self._window, title.encode("utf8"))


class MemoryWindow(BaseDebugWindow):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.NCOLS = 60
        self.NROWS = 36
        self.shift_down = False
        self.start_address = 0x0000
        self.bg_color = [0xFF, 0xFF, 0xFF]
        self.fg_color = [0x00, 0x00, 0x00]

        self._text_buffer_raw = array("B", [0x20] * (self.NROWS * self.NCOLS))
        self.text_buffer = memoryview(self._text_buffer_raw).cast("B", shape=(self.NROWS, self.NCOLS))
        self.write_border()
        self.write_addresses()

        font_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), "font.txt")
        with open(font_path) as font_file:
            font_lines = font_file.readlines()
        font_blob = "".join(line.strip() for line in font_lines[font_lines.index("BASE64DATA:\n") + 1:])
        font_bytes = zlib.decompress(b64decode(font_blob.encode()))

        self.fbuf, self.fbuf0, self.fbuf_p = make_buffer(8, 16 * 256)
        for y, b in enumerate(font_bytes):
            for x in range(8):
                self.fbuf0[y, x] = 0xFFFFFFFF if ((0x80 >> x) & b) else 0x00000000

        self.font_texture = sdl2.SDL_CreateTexture(
            self._sdlrenderer, sdl2.SDL_PIXELFORMAT_RGBA32, sdl2.SDL_TEXTUREACCESS_STATIC, 8, 16 * 256
        )

        sdl2.SDL_UpdateTexture(self.font_texture, None, self.fbuf_p, 4 * 8)
        sdl2.SDL_SetTextureBlendMode(self.font_texture, sdl2.SDL_BLENDMODE_BLEND)
        sdl2.SDL_SetTextureColorMod(self.font_texture, *self.fg_color)
        sdl2.SDL_SetRenderDrawColor(self._sdlrenderer, *self.bg_color, 0xFF)

        # Persistent to make Cython happy...
        self.src = sdl2.SDL_Rect(0, 0, 8, 16)
        self.dst = sdl2.SDL_Rect(0, 0, 8, 16)

    def write_border(self):
        for x in range(self.NCOLS):
            self.text_buffer[0][x] = 0xCD
            self.text_buffer[2][x] = 0xCD
            self.text_buffer[self.NROWS - 1][x] = 0xCD

        for y in range(3, self.NROWS):
            self.text_buffer[y][0] = 0xBA
            self.text_buffer[y][9] = 0xB3
            self.text_buffer[y][self.NCOLS - 1] = 0xBA

        self.text_buffer[0][0] = 0xC9
        self.text_buffer[1][0] = 0xBA
        self.text_buffer[0][self.NCOLS - 1] = 0xBB
        self.text_buffer[1][self.NCOLS - 1] = 0xBA

        self.text_buffer[2][0] = 0xCC
        self.text_buffer[2][9] = 0xD1
        self.text_buffer[2][self.NCOLS - 1] = 0xB9

        self.text_buffer[self.NROWS - 1][0] = 0xC8
        self.text_buffer[self.NROWS - 1][9] = 0xCF
        self.text_buffer[self.NROWS - 1][self.NCOLS - 1] = 0xBC

    def write_addresses(self):
        header = (f"Memory from 0x{self.start_address:04X} "
                  f"to 0x{self.start_address+0x3FF:04X}").encode("cp437")
        for x in range(28):
            self.text_buffer[1][x + 2] = header[x]
        for y in range(32):
            addr = f"0x{self.start_address + (0x20*y):04X}".encode("cp437")
            for x in range(6):
                self.text_buffer[y + 3][x + 2] = addr[x]

    def write_memory(self):
        for y in range(32):
            for x in range(16):
                mem = self.mb.getitem(self.start_address + 16*y + x)
                a = hex(mem)[2:].zfill(2).encode("cp437")
                self.text_buffer[y + 3][3*x + 11] = a[0]
                self.text_buffer[y + 3][3*x + 12] = a[1]

    def render_text(self):
        for y in range(self.NROWS):
            self.draw_text(0, 16 * y, self.text_buffer[y])

    def draw_text(self, x, y, text):
        self.dst.x = x
        self.dst.y = y
        for i, c in enumerate(text):
            if not 0 <= c < 256:
                logger.warn(f"Invalid character {c} in {bytes(text).decode('cp437')}") # This may error too...
                c = 0
            self.src.y = 16 * c
            if self.dst.x > self.width - 8:
                logger.warn(f"Text overrun while printing {bytes(text).decode('cp437')}")
                break
            sdl2.SDL_RenderCopy(self._sdlrenderer, self.font_texture, self.src, self.dst)
            self.dst.x += 8

    def post_tick(self):
        sdl2.SDL_RenderClear(self._sdlrenderer)
        self.write_memory()
        self.render_text()
        sdl2.SDL_RenderPresent(self._sdlrenderer)

    def _scroll_view(self, movement):
        self.start_address += movement
        self.start_address = max(0, min(self.start_address, 0x10000 - 0x400))
        # if self.start_address + 0x400 > 0x10000:
        #     self.start_address = 0x10000 - 0x400
        # if self.start_address < 0:
        #     self.start_address = 0
        self.write_addresses()

    def handle_events(self, events):
        events = BaseDebugWindow.handle_events(self, events)

        for event in events:
            # j - Next 256 bytes
            if event == WindowEvent.DEBUG_MEMORY_SCROLL_DOWN:
                if self.shift_down:
                    self._scroll_view(0x1000)
                else:
                    self._scroll_view(0x100)
            # k - Last 256 bytes
            elif event == WindowEvent.DEBUG_MEMORY_SCROLL_UP:
                if self.shift_down:
                    self._scroll_view(-0x1000)
                else:
                    self._scroll_view(-0x100)
            # shift tracking
            elif event == WindowEvent.MOD_SHIFT_ON:
                self.shift_down = True
            elif event == WindowEvent.MOD_SHIFT_OFF:
                self.shift_down = False
            elif event == WindowEvent._INTERNAL_MOUSE:
                # Scrolling
                if event.window_id == self.window_id and event.mouse_x == -1 and event.mouse_y == -1:
                    self._scroll_view(event.mouse_scroll_y * -0x100)

        return events
